import numpy as np
import matplotlib.pyplot as plt
from tmeasures import MeasureResult
import tmeasures as tm
from pathlib import Path



def get_limit(m:MeasureResult, op_code:str):
    ops={"max":np.nanmax,"min":np.nanmin}
    op = ops[op_code]
    # TODO ignore array if its all inf
    vals = [l[l!=np.inf] for l in m.layers]
    vals = [l for l in vals if len(l)>0]
    vals = np.array([op(l) for l in vals])
    return op(vals)


def plot_heatmap(m:MeasureResult, vmin=None, vmax=None):

    for i,l in enumerate(m.layers):
        d=len(l.shape)
        if d>1:
            dims = tuple(range(1,d))
            m.layers[i]=np.nanmean(l,axis=dims)

    if vmax is None: vmax = get_limit(m, "max")
    if vmin is None:
        vmin = get_limit(m, "min")
        if vmin > 0:
            vmin = 0

    n = len(m.layer_names)

    f, axes = plt.subplots(1, n, dpi=150, squeeze=False)
    mappable=None
    for i, (activation, name) in enumerate(zip(m.layers, m.layer_names)):
        ax = axes[0,i]
        ax.axis("off")
        activation = activation[:, np.newaxis]
        mappable = ax.imshow(activation,vmin=vmin,vmax=vmax,cmap='inferno',aspect="auto")

        if n<40:
            if len(name)>7:
                name=name[:6]+"."
            ax.set_title(name, fontsize=4,rotation = 45)
    f.subplots_adjust(right=0.8)
    cbar_ax = f.add_axes([0.85, 0.15, 0.05, 0.7])
    cbar=f.colorbar(mappable, cax=cbar_ax, extend='max')
    cbar.cmap.set_over('green')
    cbar.cmap.set_bad(color='gray')
    return f






import torch
import tmeasures as tm
from torch.utils.data import Dataset

from . import ActivationsModule
from .. import MeasureResult, StratifiedMeasureResult
import abc
import typing
from typing import List

ActivationsByLayer = [torch.Tensor]


class PyTorchMeasureOptions:
    def __init__(self, batch_size=32, num_workers=0, verbose=True, model_device="cpu", measure_device="cpu",
                 data_device="cpu"):
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.verbose = verbose
        self.model_device = model_device
        self.measure_device = measure_device
        self.data_device = data_device
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(b={self.batch_size},workers={self.num_workers},devices[model,measure,data]=[{self.model_device,self.measure_device,self.data_device}])"


STRowIterator = typing.Union[typing.Iterable[torch.Tensor], typing.Sized]
STMatrixIterator = typing.Union[typing.Iterable[STRowIterator], typing.Sized]


class PyTorchLayerMeasure:

    @abc.abstractmethod
    def eval(self, iterator: STMatrixIterator,layer_name:str) -> torch.Tensor:
        pass

    @abc.abstractmethod
    def generate_result(self, layer_results: ActivationsByLayer, layer_names: List[str]):
        pass


class PyTorchMeasureResult(tm.MeasureResult):

    def numpy(self):
        self.layers = [l.cpu().numpy() for l in self.layers]
        return self


class PyTorchMeasure(tm.Measure):
    def __repr__(self):
        return f"{self.abbreviation()}"

    @abc.abstractmethod
    def eval(self, dataset: Dataset, transformations: tm.TransformationSet, model: ActivationsModule,
             o: PyTorchMeasureOptions) -> PyTorchMeasureResult:
        '''

        '''
        pass

    def eval_stratified(self, datasets: typing.List[Dataset], transformations: tm.TransformationSet,
                        model: ActivationsModule, o: PyTorchMeasureOptions,
                        labels: typing.List[str]) -> StratifiedMeasureResult:
        '''
        Calculate the `variance_measure` for each class separately
        Also calculate the average stratified `variance_measure` over all classes
        '''
        results_per_set = [self.eval(dataset,transformations,model,o) for dataset in datasets]
        stratified_measure_layers = self.mean_result(results_per_set)
        stratified_measure_layers = [l.cpu().numpy() for l in stratified_measure_layers]

        layer_names = results_per_set[0].layer_names

        return StratifiedMeasureResult(stratified_measure_layers, layer_names, self, results_per_set, labels)

    def mean_result(self, measure_results: typing.List[MeasureResult]) -> ActivationsByLayer:
        # calculate the mean activation of each unit in each layer over the datasets
        results = [r.layers for r in measure_results]
        # results is a list (classes) of list layers)
        # turn it into a list (layers) of lists (classes)
        layer_class_vars = [list(i) for i in zip(*results)]
        # compute average result of each layer over datasets
        layer_vars = [sum(layer_values) / len(layer_values) for layer_values in layer_class_vars]
        return layer_vars



from tmeasures.pytorch.activations_iterator import PytorchActivationsIterator

class PyTorchMeasureByLayer(PyTorchMeasure):
    
    def __init__(self,layer_measure:PyTorchLayerMeasure) -> None:
        super().__init__()
        self.layer_measure=layer_measure

    def eval(self, dataset: Dataset, transformations: tm.TransformationSet, model: ActivationsModule,
             o: PyTorchMeasureOptions) -> PyTorchMeasureResult:
        dataset2d = tm.pytorch.dataset2d.TransformationSampleDataset(dataset, transformations, device=o.data_device)
        iterator = PytorchActivationsIterator(model, dataset2d, o)
        results = iterator.evaluate(self.layer_measure)
        return PyTorchMeasureResult(results, model.activation_names(), self)


import numpy as np
import matplotlib.pyplot as plt
from .. import MeasureResult
from pathlib import Path
from typing import List


def relative_error(reference: MeasureResult, x: MeasureResult):
    error = 0
    n = 0
    for reference_layer, layer in zip(reference.layers, x.layers):
        # comparing with e_rel = |(x-y)| / (|x|+|y|)
        # Note that if (|x|-|y|) ~= 0 then (x-y) ~= 0,
        # So we set e_rel = 0 to avoid division by 0
        num = np.abs(reference_layer - layer)
        den =  np.abs(reference_layer)+np.abs(layer)
        error_layer = num/den
        error_layer[np.isclose(den,0.0)]=0
        error += error_layer.sum()
        n += np.prod(layer.shape)
    return error / n


def get_relative_errors(results: 'np.ndarray[MeasureResult]', reference: MeasureResult):
    # Calcuate the relative error between the reference and each value of the result array
    def compare(r: MeasureResult): return relative_error(reference, r)
    compare_v = np.vectorize(compare)
    return compare_v(results)



def plot_relative_error_heatmap(heatmap, 
                                labels_samples: List[str], labels_transformations: List[str]):
    # plot a heatmap of relative errors between the MeasureResults in `results` and `reference`
    
    
    f, ax = plt.subplots(dpi=300)
    im = ax.imshow(heatmap, cmap="magma",vmin=0,vmax=1)
    # ax.set_xticklabels(labels_samples)
    # ax.set_yticklabels(labels_transformations)
    plt.xticks(range(len(labels_samples)),labels_samples)
    plt.yticks(range(len(labels_transformations)),labels_transformations)
    ax.set_xlabel("Samples")
    ax.set_ylabel("Transformations")
    cbar = ax.figure.colorbar(im, ax=ax)

    for i in range(heatmap.shape[0]):
        for j in range(heatmap.shape[1]):
            value_str = f"{heatmap[i, j]:.3f}"
            text = ax.text(j, i, value_str,
                           ha="center", va="center", color="#8aff66")




#from __future__ import annotations
import abc
from typing import List
from tmeasures.transformation import TransformationSet

import numpy as np


class ActivationsIterator(abc.ABC):
    """
    .. 
        Iterate over the ST matrix of activations of a network, varying the samples and transformations.
        The iteration is in both orders:
        1) transformations_first:  column indexes first, then row indices indices
                For transformation 0, get all samples (col 1, all rows), then (col 2, all rows)
                The iteration over rows in done by batches.
        2) samples_first: row indexes first, then column
                Same as transformations_first but transposed.

    """

    @abc.abstractmethod
    def get_transformations(self)->TransformationSet:
        pass

    @abc.abstractmethod
    def transformations_first(self):
        pass

    @abc.abstractmethod
    def samples_first(self):
        pass

    @abc.abstractmethod
    def layer_names(self) -> List[str]:
        pass

    @abc.abstractmethod
    def get_inverted_activations_iterator(self) -> 'ActivationsIterator':
        pass

    @abc.abstractmethod
    def get_both_iterator(self) -> 'ActivationsIterator':
        pass

    @abc.abstractmethod
    def get_normal_activations_iterator(self) -> 'ActivationsIterator':
        pass

    def row_from_iterator(self,transformation_activations_iterator):
        '''
        Get a row of the ST matrix from the :param transformation_activations_iterator
        :return: row of the ST matrix with the  activations for all the transformations of sample, and also the transformed samples
        '''
        activations=[[] for i in range(len(self.layer_names()))]
        x_transformed=[]
        for  x_transformed_batch,activations_batch in transformation_activations_iterator:
            x_transformed.append(x_transformed_batch)
            for i, layer_activations in enumerate(activations_batch):
                activations[i].append(layer_activations)
        activations = [ np.concatenate(layer_activations,axis=0) for layer_activations in activations]
        x_transformed = np.concatenate(x_transformed,axis=0)
        return activations,x_transformed






from __future__ import annotations
from typing import List,Tuple,Sized,Iterable,Iterator
import numpy as np
import torch
import abc

class Transformation:

    @abc.abstractmethod
    def __call__(self, x):
        pass

    def parameters(self):
        return np.array([])

class InvertibleTransformation:

    @abc.abstractmethod
    def inverse(self) -> InvertibleTransformation:
        pass

class TransformationSet(list,Sized, Iterable[Transformation]):

    def __init__(self,members):
        super().__init__(members)

    @abc.abstractmethod
    def id(self):
        pass

    @abc.abstractmethod
    def valid_input(self,shape:Tuple[int, ])->bool:
        pass

    @abc.abstractmethod
    def copy(self)->'TransformationSet':
        pass

    @abc.abstractmethod
    def parameter_range(self):
        pass


class IdentityTransformation(InvertibleTransformation):
    def __call__(self, x):
        return x

    def inverse(self) -> InvertibleTransformation:
        return self





from typing import List, Tuple
from .base import NumpyMeasure
from ..measure import ActivationsByLayer
from .. import MeasureResult
from .activations_iterator import ActivationsIterator
from .stats_running import RunningMeanWelford
import scipy.stats

class ANOVAInvariance(NumpyMeasure):
    # alpha = degree of confidence
    # Typically 0.90, 0.95, 0.99
    def __init__(self, alpha:float=0.99,bonferroni:bool=True):
        super().__init__()
        self.anova_f_measure=ANOVAFInvariance()
        assert(alpha>0)
        assert (alpha <1)
        self.alpha=alpha
        self.bonferroni=bonferroni

    def __repr__(self):
        return f"ANOVAI(α={self.alpha},b={self.bonferroni})"

    def eval(self, activations_iterator: ActivationsIterator,verbose=False) -> MeasureResult:
        f_result=self.anova_f_measure.eval(activations_iterator)
        d_b=f_result.extra_values["d_b"]
        d_w = f_result.extra_values["d_w"]
        n_layers=len(f_result.layer_names)
        layers=f_result.layers
        if self.bonferroni:
            # use adjusted alphas to calculate critical values
            critical_values = []
            for layer in layers:
                adjusted_alpha=self.alpha/layer.size
                critical_values.append(scipy.stats.f.ppf(adjusted_alpha,dfn=d_b,dfd=d_w))
        else:
            critical_values = [scipy.stats.f.ppf(self.alpha,dfn=d_b,dfd=d_w)]*n_layers

        f_critical=zip(layers,critical_values)
        layers=[ f>critical_value for (f,critical_value) in f_critical]
        return MeasureResult(layers,f_result.layer_names,self,f_result.extra_values)

    def name(self):
        return "ANOVA"
    def abbreviation(self):
        return self.name()

class ANOVAFInvariance(NumpyMeasure):
    def __init__(self):
        super().__init__()


    def __repr__(self):
        return f"{self.__class__.__name__}()"


    def eval(self,activations_iterator:ActivationsIterator,verbose=False)->MeasureResult:

        # calculate mean(X_t)
        u_t,n_t=self.eval_means_per_transformation(activations_iterator)
        n_layers=len(activations_iterator.layer_names())
        # Calculate mean(X)
        u=self.eval_global_means(u_t,n_layers)
        # Calculate  mean_t[ (mean(X_t)-mean(X))^2], that is Var( mean(X_t) ). Normalized with T-1
        ssb, d_b=self.eval_between_transformations_ssd(u_t,u,n_t)

        # Calculate t: (mean(X_t)-X)², that is Var(X_t). Normalized with N-T
        ssw, d_w=self.eval_within_transformations_ssd(activations_iterator, u_t)
        # calculate
        f_score=self.divide_per_layer(ssb,ssw)

        return MeasureResult(f_score, activations_iterator.layer_names(), self, extra_values={"d_b":d_b, "d_w":d_w})

    def divide_per_layer(self,a_per_layer:ActivationsByLayer,b_per_layer:ActivationsByLayer)->Tuple[ActivationsByLayer,int]:
        return [a/b for (a,b) in zip(a_per_layer,b_per_layer)]

    def eval_within_transformations_ssd(self,activations_iterator:ActivationsIterator,means_per_layer_and_transformation:List[ActivationsByLayer],)->ActivationsByLayer:
            n_layers = len(activations_iterator.layer_names())

            ssdw_per_layer = [0] * n_layers
            samples_per_transformation = []
            for means_per_layer,(transformation, transformation_activations) in zip(means_per_layer_and_transformation, activations_iterator.transformations_first()):
                # calculate the variance of all samples for this transformation
                n_samples = 0
                for x, batch_activations in transformation_activations:
                    n_samples += x.shape[0]
                    for j, layer_activations in enumerate(batch_activations):
                        for i in range(layer_activations.shape[0]):
                            d=(layer_activations[i,]-means_per_layer[j])**2
                            ssdw_per_layer[j]=ssdw_per_layer[j]+d
                samples_per_transformation.append(n_samples)
            # divide by degrees of freedom
            degrees_of_freedom = (samples_per_transformation[0] - 1) * len(samples_per_transformation)
            ssdw_per_layer = [s / degrees_of_freedom for s in ssdw_per_layer]
            return ssdw_per_layer,degrees_of_freedom

    def eval_between_transformations_ssd(self,means_per_layer_and_transformation:List[ActivationsByLayer],global_means:ActivationsByLayer,samples_per_transformation:List[int])->Tuple[ActivationsByLayer,int]:
        '''

        :param means_per_transformation: has len(transformations), each item has len(layers)
        :param global_means: has len(layers)
        :return:
        '''
        n_layers=len(global_means)
        n_transformations=len(means_per_layer_and_transformation)
        ssdb_per_layer=[0]*n_layers
        for  i,means_per_layer in enumerate(means_per_layer_and_transformation):
            n=samples_per_transformation[i]
            for j in range(n_layers):
                ssdb_per_layer[j] += n* ((means_per_layer[j]-global_means[j])**2)
        degrees_of_freedom=(n_transformations-1)
        for j in range(n_layers):
            ssdb_per_layer[j]/=degrees_of_freedom
        return ssdb_per_layer,degrees_of_freedom

    def eval_global_means(self, means_per_layer_and_transformation:List[ActivationsByLayer], n_layers:int)->ActivationsByLayer:
        '''

        :param means_per_layer_and_transformation:
        :param n_layers:
        :return: The global means for each layer, averaging out the transformations
        '''
        # n_transformations = len(means_per_layer_and_transformation)
        global_means_running = [RunningMeanWelford() for i in range(n_layers)]
        for transformation_means in means_per_layer_and_transformation:
            # means_per_layer  has the means for a given transformation
            for i,layer_means in enumerate(transformation_means):
                global_means_running[i].update(layer_means)

        return [rm.mean() for rm in global_means_running]




    def eval_means_per_transformation(self,activations_iterator:ActivationsIterator)->Tuple[List[ActivationsByLayer],List[int]]:
        '''
        For all activations, calculates the mean activation value for each transformation
        param activations_iterator
        return A list of mean activation values for each activation in each layer
        The list of samples per transformation
        '''
        n_layers = len(activations_iterator.layer_names())
        means_per_transformation = []
        samples_per_transformation=[]
        for transformation, transformation_activations in activations_iterator.transformations_first():
            samples_variances_running = [RunningMeanWelford() for i in range(n_layers)]
            # calculate the variance of all samples for this transformation
            n_samples=0
            for x, batch_activations in transformation_activations:
                n_samples+=x.shape[0]
                for j, layer_activations in enumerate(batch_activations):
                    for i in range(layer_activations.shape[0]):
                        samples_variances_running[j].update(layer_activations[i,])
            samples_per_transformation.append(n_samples)
            means_per_transformation.append([rm.mean() for rm in samples_variances_running])
        return means_per_transformation,samples_per_transformation





from .dataset2d import STDataset, Dataset2D
import torch
from torch.utils.data import DataLoader
from . import ActivationsModule
from .base import PyTorchMeasureOptions, PyTorchLayerMeasure, PyTorchMeasure

from .activations_transformer import ActivationsTransformer
from ..utils.iterable_queue import IterableQueue
import queue
from .. import Transformation, InvertibleTransformation
import tqdm.auto as tqdm
import concurrent.futures

try:
    import namedthreads
    namedthreads.patch()
except ImportError: pass


from typing import List

import abc


class ActivationsTransformer(abc.ABC):

    @abc.abstractmethod
    def transform(self, activations: torch.Tensor, x: torch.Tensor, transformations: List[Transformation]) -> torch.Tensor:
        pass


class IdentityActivationsTransformer(ActivationsTransformer):
    def transform(self, activations: torch.Tensor, x: torch.Tensor, transformations: List[Transformation]) -> torch.Tensor:
        return activations

from tmeasures import logger
from concurrent.futures import ThreadPoolExecutor
import threading

class ThreadsManager:

    def __init__(self,layers:list[str],rows:int,n_batch:int,stop=False) -> None:
        self.stop=stop
        self.layers = layers
        self.qs = {l: IterableQueue(rows,maxsize=1,name=f"q({l})") for l in layers}
        self.row_qs = {l: IterableQueue(n_batch,maxsize=1,name=f"q({l}_row)") for l in layers}
    @property
    def queues(self):
        return list(self.qs.values())+list(self.row_qs.values())
    
    def cancel(self,server_future):
        logger.info("Exception raised when computing measure, shutting down all computing threads...")
        self.stop=True
        for q in self.queues:
            q.stop=True
        logger.info(f"Emptying queues to ensure server can move on and stop..")
        self.empty_all()
        # wait for server thread to finish
        concurrent.futures.wait([server_future], return_when=concurrent.futures.ALL_COMPLETED)
        # threading.Event().wait(0.1)
        logger.info("Server stopped. Pushing values to queues to ensure workers can consume and stop..")
        self.stop_all()
        logger.info("Workers stopped")
    
    def empty_all(self):
        for q in self.queues:
            try: q.get(block=False) 
            except queue.Empty: pass
 

    def stop_all(self):
        for q in self.queues:
            try: q.put(None,block=False)
            except queue.Full: pass
    

        

class PytorchActivationsIterator:

    def __init__(self, model: ActivationsModule, dataset: Dataset2D, o: PyTorchMeasureOptions,
                 activations_transformer: ActivationsTransformer = IdentityActivationsTransformer()):
        self.model = model
        self.dataset = dataset
        self.o = o
        self.activations_transformer = activations_transformer

    def check_finished(self,worker_futures,server_future,tm:ThreadsManager):
        futures = [server_future]+list(worker_futures.values())
        logger.info(f"Waiting for {len(worker_futures)} workers and server to finish...")
        done, not_done = concurrent.futures.wait(futures, return_when=concurrent.futures.FIRST_EXCEPTION)
        logger.info(f"done")
        # check if an exception was raised
        
        if server_future in done:
            server_exception = server_future.exception()
        else:
            server_exception = None
        worker_exceptions = [f.exception() if f in done else None for f in worker_futures.values()]
        worker_failure = any([not e is None for e in worker_exceptions])
        if server_exception is None and not worker_failure:
            logger.info("No exceptions found, threads finished.")
            return None
        else: # we have some exceptions
            # signal stop for server
            logger.info("Some threads failed, terminating")
            tm.cancel(server_future)
            done, not_done = concurrent.futures.wait(futures, return_when=concurrent.futures.ALL_COMPLETED)

            if not server_exception is None:
                logger.info(f"Server exception, about to re raise from main thread\n{server_exception}\n")
                raise server_exception
            else:
                for e in worker_exceptions:
                    if not e is None:
                        logger.info(f"Worker exception, about to re raise from main thread\n{e}\n thread id {threading.get_ident()}\n")
                        raise e
                    
    def feed_threads(self,tm:ThreadsManager):
         layers = self.model.activation_names()
         rows, cols = self.dataset.len0, self.dataset.len1
         
         with torch.no_grad():
                # print(f"act it starting,num workers {self.o.num_workers}:")
                for row in tqdm.trange(rows, disable=not self.o.verbose, leave=False):
                    row_dataset = self.dataset.row_dataset(row)
                    row_dataloader = DataLoader(row_dataset, batch_size=self.o.batch_size, shuffle=False, num_workers=0,pin_memory=True)
                    
                    for k, q in tm.qs.items():
                        logger.info(f"AI: putting row {row} dataloader for  layer {k}")
                        q.put(tm.row_qs[k])

                    # print(f"AI: finished putting row {row} dataloaders for all layers")
                    # for k,q in qs.items():
                    #     print(f"AI: {k}→ {q.queue.qsize()} items")
                    if tm.stop:
                            logger.info("Server thread stopping, exception detected")
                            return
                    col = 0
                    # print("col",col)
                    
                    for batch_i,x_transformed in enumerate(row_dataloader):
                        # print(f"AI: {batch_i}: moving to device {self.o.model_device}... ")
                        x_transformed = x_transformed.to(self.o.model_device,non_blocking=True)
                        # print("AI: getting activations..")
                        activations = self.model.forward_activations(x_transformed)
                        # print("AI: got activations")
                        col_to = col + x_transformed.shape[0]
                        for i, layer_activations in enumerate(activations):
                            if self.o.model_device != self.o.measure_device:
                                layer_activations=layer_activations.to(self.o.measure_device,non_blocking=True)

                            
                            transformations = self.dataset.get_transformations(row, col, col_to)
                            layer_activations = self.activations_transformer.transform(layer_activations, x_transformed,transformations)
                            # print(f"AI: act it, shape {layer_activations.shape}")
                            # print(f"AI: putting col {col} batch for layer {i} ({layers[i]})")
                            tm.row_qs[layers[i]].put(layer_activations)
                            # print(f"put {layer_activations.shape} into {layers[i]} {row_qs[layers[i]]}")
                            # Check if there's been an exception 
                            if tm.stop:
                                logger.info("Server thread stopping, exception detected")
                                return
                        col = col_to
                        # print("AI: finished row")
                    # print("AI: finished all rows")
   

    def evaluate(self, m: PyTorchLayerMeasure):
        self.queues=[]
        layers = self.model.activation_names()
        rows, cols = self.dataset.len0, self.dataset.len1
        prefix = f"{m.__class__.__name__}_"
        logger.info(f"Main thread {threading.get_ident()}")
        # calculate number of batches per row
        n_batch = cols // self.o.batch_size 
        if (cols % self.o.batch_size) >0:
            n_batch+=1

        
        with ThreadPoolExecutor(max_workers=len(layers)+1,thread_name_prefix=prefix) as executor:
            tm = ThreadsManager(layers,rows,n_batch,stop=False)
            logger.info("starting server thread")
            server_thread = executor.submit(self.feed_threads,tm)
            logger.info("starting worker threads")
            worker_threads = {l:executor.submit(m.eval,q,l) for l,q in tm.qs.items()}
            
            r = self.check_finished(worker_threads,server_thread,tm)
            
            if r is None:
                return [t.result() for t in worker_threads.values()]
            else:
                raise r 



import matplotlib
import matplotlib.pyplot as plt
from .. import MeasureResult


def discrete_colormap(n:int=16,base_colormap="rainbow",):
    colors = plt.cm.get_cmap(base_colormap, n)(range(n))
    cm = matplotlib.colors.ListedColormap(colors)
    return cm


def default_discrete_colormap():
    return plt.cm.get_cmap("Set1")

def get_sequential_colors(values):
    cmap= plt.cm.get_cmap("plasma",len(values))
    colors = cmap(values)
    return colors

from .layers import plot_average_activations_same_model,plot_average_activations_different_models,plot_average_activations,scatter_same_model
from .heatmaps import plot_heatmap
from .features import plot_invariant_feature_maps
from .sample_size import plot_relative_error_heatmap,get_relative_errors





from typing import Any, List
from . import ActivationsModule
from torch import nn
import torch
class SequentialWithIntermediates(nn.Sequential, ActivationsModule):
    def __init__(self,*args):
        super(SequentialWithIntermediates, self).__init__(*args)

    def forward_intermediates(self,input_tensor)->(List[torch.Tensor]):
        submodules=self._modules.values()
        if len(submodules)==0:
            return  input_tensor,[input_tensor]

        outputs=[]
        for module in submodules:
            if isinstance(module, ActivationsModule):
                intermediates=module.forward_activations(input_tensor)
                outputs+=(intermediates)
            else:
                input_tensor= module(input_tensor)
                outputs.append(input_tensor)
        return outputs

    def activation_names(self)->List[str]:
        submodules = self._modules.values()
        if len(submodules) == 0:
            return ["identity"]
        if len(submodules) == 1:
            module = list(submodules)[0]

            if isinstance(module, ActivationsModule):
                return ["0_"+name for name in module.activation_names()]
            else:
                name = module.__class__.__name__
                return [self.abbreviation(name)]

        # len(submodules)>1
        outputs = []
        index=0

        for module in submodules:
            if isinstance(module, ActivationsModule):
                index += 1
                module_name=self.abbreviation(module.__class__.__name__)
                names=[f"{module_name}{index}_{name}" for name in module.activation_names()]
                outputs +=names
            else:
                name=module.__class__.__name__
                if name.startswith("Conv") or name.startswith("Linear"):
                    index += 1  # conv and fc layers increase index
                name=f"{index}{self.abbreviation(name)}"
                outputs.append(name)
        return outputs

    def abbreviation(self, name:str)->str:
        if name.startswith("Conv"):
            name = "c"
        elif name.startswith("BatchNorm"):
            name = "bn"
        elif name.startswith("ELU"):
            name = "elu"
        elif name.startswith("ReLU"):
            name= "relu"
        elif name.startswith("Linear"):
            name = "fc"
        elif name.startswith("Add"):
            name = "+"
        elif "Softmax" in name:
            name="sm"
        elif name == "Sequential":
            name = ""
        elif name == "SequentialWithIntermediates":
            name = ""
        elif name == "Block":
            name = "b"
        return name
    



import scipy.spatial.distance
import numpy as np
from abc import ABC,abstractmethod

class ActivationsTransformation(ABC):

    @abstractmethod
    def apply(self,activations:np.ndarray):
        pass

class DistanceAggregation(ActivationsTransformation):

    def __init__(self, normalize:bool, keep_shape:bool, distance="euclidean"):
        self.normalize=normalize
        self.keep_shape=keep_shape
        self.distance=distance


    def __repr__(self):
        return f"DA(normalize={self.normalize},keep_shape={self.keep_shape},distance={self.distance})"

    def normalize_activations(self,x:np.ndarray):
        c,n,d=x.shape
        if self.normalize and d>1:
            for i in range(c):
                for j in range(n):
                    x[i,j,:]/=np.linalg.norm(x[i,j,:])

    def apply(self,x:np.ndarray):
        x = self.convert_to_cnd_format(x)
        # x has size (c,n,d), where c is the feature dimension
        self.normalize_activations(x)
        return self.aggregate_distances(x)

    def convert_to_cnd_format(self,x:np.ndarray):
        l = len(x.shape)
        # Convert x to shape (n, features, dim_features)
        if l > 2:
            n, features, = x.shape[0:2]
            if self.keep_shape:
                # consider dims 2+ as a single dim
                # dim 0 is the number
                x = x.reshape((n, features, -1))

            else:
                # consider every element of the feature map as a distinct activation
                x = x.reshape((n, -1, 1))
        elif l == 2:
            n,c=x.shape
            x = x.reshape((n, c, 1))
        else:
            raise ValueError(f"Activation shape not supported {x.shape}")
        # ncd to cnd
        x = x.transpose((1, 0, 2))
        x = np.ascontiguousarray(x)
        return x

    def aggregate_distances(self, x: np.ndarray):
        c,n, d = x.shape
        results = np.zeros(c)
        for i in range(c):
            sample = x[i, :,:]
            dm = scipy.spatial.distance.pdist(sample, self.distance)
            results[i]= dm.mean()
        return results





from typing import List
from tmeasures import MeasureResult
from .activations_iterator import ActivationsIterator
from . import NumpyMeasure
from .quotient import divide_activations

from multiprocessing import Queue
from .multithread.multithreaded_layer_measure import LayerMeasure,PerLayerMeasure,ActivationsOrder
import numpy as np
from .stats_running import RunningMeanAndVarianceWelford,RunningMeanWelford
from scipy.stats import norm
from tqdm import tqdm

default_alpha=0.99
default_sign=1

class GoodfellowNormalGlobalInvariance(NumpyMeasure):
    thresholds_key="thresholds"

    def __repr__(self):
        return f"{self.__class__.__name__}(gp={self.alpha})"

    # def name(self):
    #     return "GoodfellowNormalGlobal"
    #
    # def abbreviation(self):
    #     return "GGN"

    def __init__(self, alpha:float=default_alpha, sign:int=default_sign):
        super().__init__()
        self.alpha = alpha
        self.sign=sign

    def eval(self,activations_iterator: ActivationsIterator,verbose=False)->MeasureResult:
        running_means = [RunningMeanAndVarianceWelford() for i in activations_iterator.layer_names()]

        for transformation, samples_activations_iterator in tqdm(activations_iterator.transformations_first(),disable=not verbose):
            for x, batch_activations in samples_activations_iterator:
                for j, activations in enumerate(batch_activations):
                    if self.sign != 1: activations *= self.sign
                    running_means[j].update_all(activations)

        stds  = [running_mean.std() for running_mean in running_means]
        means = [running_mean.mean() for running_mean in running_means]
        original_shapes = [mean.shape for mean in means]
        means = [mean.reshape(mean.size) for mean in means]
        stds =  [std.reshape(std.size) for std in stds]
    # calculate the threshold values (approximately)
        thresholds=[np.zeros(mean.size) for mean in means]

        for i,(mean,std) in enumerate(zip(means,stds)):
            for j,(mu,sigma) in enumerate(zip(mean,std)):
                if sigma>0:
                    t=norm.ppf(self.alpha,loc=mu,scale=sigma)
                else:
                    t=mu
                thresholds[i][j]=t
    #thresholds = mean+2*std
        thresholds=[threshold.reshape(original_shape) for threshold,original_shape in zip(thresholds,original_shapes)]
        # set g(i) equal to the activations_percentage
        layers_g= [np.zeros_like(threshold) + (1-self.alpha) for threshold in thresholds]

        return MeasureResult(layers_g, activations_iterator.layer_names(), self,extra_values={self.thresholds_key:thresholds})

class GoodfellowNormalLocalInvariance(NumpyMeasure):

    def __init__(self, thresholds:List[np.ndarray],sign:int=default_sign):
        super().__init__()
        self.thresholds = thresholds
        self.sign=sign

    def __repr__(self):
        return f"{self.__class__.__name__}"

    def eval(self,activations_iterator: ActivationsIterator,verbose=False)->MeasureResult:
        running_means = [RunningMeanWelford() for i in activations_iterator.layer_names()]

        for x,transformation_activations  in tqdm(activations_iterator.samples_first(),disable=not verbose):
            for x_transformed, activations in transformation_activations:
                for i, layer_activations in enumerate(activations):

                    if self.sign != 1:
                        layer_activations *= self.sign

                    activated:np.ndarray = (layer_activations > self.thresholds[i]) * 1.0
                    # print(activated.shape,activated.min(),activated.max(),activated.dtype)
                    if np.any(activated<0):
                        print(activated)
                    running_means[i].update_all(activated)

        layers_l = [m.mean() for m in running_means]

        return MeasureResult(layers_l, activations_iterator.layer_names(), self)





class GoodfellowNormalInvariance(NumpyMeasure):
    g_key="global"
    l_key="local"


    def __init__(self, alpha=0.99, sign=1):
        assert sign in [1,-1]
        super().__init__()
        self.alpha=alpha
        self.sign=sign

    def eval(self,activations_iterator:ActivationsIterator,verbose=False):
        self.g = GoodfellowNormalGlobalInvariance(self.alpha, self.sign)
        g_result = self.g.eval(activations_iterator,verbose)
        thresholds = g_result.extra_values[GoodfellowNormalGlobalInvariance.thresholds_key]
        self.l = GoodfellowNormalLocalInvariance(thresholds, self.sign)
        l_result = self.l.eval(activations_iterator,verbose)

        ratio = divide_activations(l_result.layers,g_result.layers)
        extra = {self.g_key:g_result,self.l_key:l_result}

        return MeasureResult(ratio, activations_iterator.layer_names(), self,extra_values=extra)


    def __repr__(self):
        return f"{self.__class__.__name__}(gp={self.alpha})"

    def name(self):
        return "Goodfellow Normal Invariance"

    def abbreviation(self):
        return "GFN"





from torch.utils.data import Dataset, IterableDataset, DataLoader
import abc
from .transformations import PyTorchTransformationSet
import torch

class Dataset2D(Dataset):

    @property
    @abc.abstractmethod
    def len0(self):
        pass

    @property
    @abc.abstractmethod
    def len1(self):
        pass

    @property
    @abc.abstractmethod
    def T(self):
        pass

    def len(self):
        return self.len0 * self.len1

    def __getitem__(self, idx):
        if len(idx) == 1:
            i, j = idx // self.len1, idx % self.len1
            return self.getitem2d(i, j)
        elif len(idx) == 2:
            i, j = idx
            return self.getitem2d(i, j)
        else:
            raise ValueError(f"Invalid index: {idx}.")

    @abc.abstractmethod
    def getitem2d(self, i, j):
        pass

    def row_dataset(self, row: int):
        return RowDataset(self,row)

    @abc.abstractmethod
    def get_transformations(self, row:int, col_from:int, col_to:int):
        pass


class STDataset(Dataset2D):

    def __init__(self, dataset: Dataset, transformations: PyTorchTransformationSet,device=torch.device("cpu")):
        """
        @param dataset: Non iterable dataset from which to draw samples
        @param transformations: set of transformations to apply to samples
        """

        if isinstance(dataset, IterableDataset):
            raise ValueError(
                f"{IterableDataset} not supported; must specify a map-style dataset (https://pytorch.org/docs/stable/data.html#dataset-types)")
        self.dataset = dataset
        self.transformations = transformations
        self.device=device

    def len_dataset(self):
        return len(self.dataset)

    def len_transformations(self):
        return len(self.transformations)


class SampleTransformationDataset(STDataset):

    def T(self):
        return TransformationSampleDataset(self.dataset, self.transformations)

    def getitem2d(self, i, j):
        s = self.dataset[i].to(self.device)
        t = self.transformations[j]
        return t(s)

    @property
    def len0(self):
        return self.len_dataset()

    @property
    def len1(self):
        return self.len_transformations()

    def get_transformations(self, row: int, col_from: int, col_to: int):
        return self.transformations[col_from:col_to]

class TransformationSampleDataset(STDataset):

    def getitem2d(self, i, j):
        t = self.transformations[i]
        s = self.dataset[j].to(self.device)
        return t(s)

    def T(self):
        return SampleTransformationDataset(self.dataset, self.transformations)

    @property
    def len1(self):
        return self.len_dataset()

    @property
    def len0(self):
        return self.len_transformations()

    def get_transformations(self, row: int, col_from: int, col_to: int):
        return [self.transformations[row]]*(col_to-col_from)

class RowDataset(Dataset):

    def __init__(self, d: Dataset2D, row: int):
        self.d = d
        self.row = row

    def __getitem__(self, item):
        return self.d.getitem2d(self.row, item)

    def __len__(self):
        return self.d.len1






from ..stats_running import RunningMeanAndVarianceWelford,RunningMeanAndVarianceWelford
from enum import Enum
import numpy as np

class MeasureFunction(Enum):
    var = "var"
    std = "std"
    mean = "mean"

    def apply_running(self, rm:RunningMeanAndVarianceWelford):
        if self == MeasureFunction.var:
            return rm.var()
        elif self == MeasureFunction.std:
            return rm.std()
        elif self == MeasureFunction.mean:
            return rm.mean()
        else:
            raise ValueError(f"Unsupported measure function {self.measure_function}")

    def apply(self, activations):
        if (self == MeasureFunction.var or self == MeasureFunction.std) and activations.shape[0]<=1:
            return np.zeros_like(activations[0,:])
        functions = {
            MeasureFunction.var: lambda x: np.var(x, axis=0,ddof=1)
            , MeasureFunction.std: lambda x: np.std(x, axis=0,ddof=1)
            , MeasureFunction.mean: lambda x: np.mean(x, axis=0)
        }
        function = functions[self]
        return function(activations)






from tmeasures import MeasureResult
from ..activations_iterator import ActivationsIterator
from ..base import NumpyMeasure
import tmeasures as tm
from multiprocessing import Queue
from .multithreaded_layer_measure import LayerMeasure,PerLayerMeasure,ActivationsOrder
import numpy as np
from ..stats_running import RunningMeanAndVarianceWelford, RunningMean
from scipy.stats import norm

class GlobalVarianceNormal(LayerMeasure):

    def __init__(self, id:int, name:str, alpha:float, sign:int):
        super().__init__(id,name)
        self.alpha = alpha
        self.sign=sign

    def eval(self,q:Queue,inner_q:Queue):
        running_mean = RunningMeanAndVarianceWelford()
        for transformation in self.queue_as_generator(q):
            i=0
            for activations in self.queue_as_generator(inner_q):
                if self.sign != 1:
                    activations *= self.sign
                running_mean.update_all(activations)

        std=running_mean.std()
        mean = running_mean.mean()
        original_shape = mean.shape
        mean=mean.reshape(mean.size)
        std = std.reshape(std.size)
        # calculate the threshold values (approximately)
        thresholds=np.zeros(mean.size)
        for i,(mu,sigma) in enumerate(zip(mean,std)):
            t=norm.ppf(self.alpha,loc=mu,scale=sigma)
            thresholds[i]=t
        #thresholds = mean+2*std
        thresholds=thresholds.reshape(original_shape)
        # set g(i) equal to the activations_percentage
        g= np.zeros_like(thresholds) + (1-self.alpha)
        return g,thresholds

class LocalVarianceNormal(LayerMeasure):

    def __init__(self, id:int, name:str, threshold:float,sign:int):
        super().__init__(id,name)
        self.threshold = threshold
        self.sign=sign

    def eval(self,q:Queue,inner_q:Queue):
        running_mean = RunningMean()
        # activation_sum=0
        n=0
        for transformation in self.queue_as_generator(q):
            for activations in self.queue_as_generator(inner_q):
                if self.sign != 1:
                    activations *= self.sign
                activated = (activations > self.threshold) * 1.0
                running_mean.update_all(activated)

        return running_mean.mean()




class GlobalFiringRateNormalMeasure(PerLayerMeasure):
    def __init__(self,activation_percentage:float,sign:int):
        super().__init__(ActivationsOrder.TransformationsFirst)
        self.activation_percentage:float=activation_percentage
        self.sign:int=sign

    def __repr__(self):
        return f"G()"

    def generate_layer_measure(self, i:int, name:str) -> LayerMeasure:
        return GlobalVarianceNormal(i, name, self.activation_percentage, self.sign)

    def generate_result_from_layer_results(self, results_tresholds, names):
        results_tresholds, tresholds = zip(*results_tresholds)
        return MeasureResult(results_tresholds, names, self, extra_values={"thresholds":tresholds})

class LocalFiringRateNormalMeasure(PerLayerMeasure):
    def __init__(self,thresholds:[np.ndarray],sign:int):
        super().__init__(ActivationsOrder.SamplesFirst)
        self.sign=sign
        self.thresholds=thresholds

    def __repr__(self):
        return f"L()"

    def generate_layer_measure(self, i:int, name:str) -> LayerMeasure:
        return LocalVarianceNormal(i, name, self.thresholds[i], self.sign)

class GoodfellowNormalMeasure(NumpyMeasure):


    def __init__(self, alpha=0.99, sign=1):
        assert sign in [1,-1]
        super().__init__()
        self.alpha=alpha
        self.sign=sign

    def eval(self,activations_iterator:ActivationsIterator):
        self.g = GlobalFiringRateNormalMeasure(self.alpha, self.sign)
        g_result = self.g.eval(activations_iterator)
        self.thresholds = g_result.extra_values["thresholds"]
        self.l = LocalFiringRateNormalMeasure(self.thresholds, self.sign)
        l_result = self.l.eval(activations_iterator)

        ratio = tm.divide_activations(l_result.layers,g_result.layers)
        return MeasureResult(ratio, activations_iterator.layer_names(), self)


    def __repr__(self):
        return f"GoodfellowNormal(gp={self.alpha})"

    def name(self):
        return "Goodfellow Normal"
    def abbreviation(self):
        return "GFN"



"""
The :mod:`.pytorch` module contains implementations of the measures for the PyTorch framework, as well as some utility functions.

A :class:`.PyTorchMeasure` can be evaluated with a PyTorch :class:`.ActivationsModule`, dataset and :class:`.PyTorchTransformationSet`. The evaluation of the measure returns a :class:`.PyTorchMeasureResult`

Variance-Based Invariance Measures
------------------------
* :class:`.SampleVarianceInvariance`
* :class:`.TransformationVarianceInvariance`
* :class:`.NormalizedVarianceInvariance`

Other Invariance Measures
------------------------
* :class:`.GoodfellowInvariance`

Variance-Based Same-Equivariance Measures
------------------------
* :class:`.SampleVarianceSameEquivariance`
* :class:`.TransformationVarianceSameEquivariance`
* :class:`.NormalizedVarianceSameEquivariance`
"""

from .model import ActivationsModule,AutoActivationsModule
from .base import PyTorchMeasure,PyTorchMeasureOptions,ActivationsByLayer,PyTorchMeasureResult
from .transformations import PyTorchTransformationSet,PyTorchTransformation

from .activations_iterator_base import PytorchActivationsIterator,InvertedPytorchActivationsIterator,BothPytorchActivationsIterator,NormalPytorchActivationsIterator
from .measure_transformer import MeasureTransformation,NoTransformation,AverageFeatureMaps
from .util import SequentialWithIntermediates
from . import dataset2d

from .layer_measures import Variance

from .variance_invariance import SampleVarianceInvariance,TransformationVarianceInvariance,NormalizedVarianceInvariance
from .variance_sameequivariance import  SampleVarianceSameEquivariance,TransformationVarianceSameEquivariance, NormalizedVarianceSameEquivariance
from .goodfellow import GoodfellowInvariance,NormalPValueThreshold,PercentActivationThreshold


import abc
import torch

class TransformationAdapter:

    @abc.abstractmethod
    def pre_adapt(self,x):
        pass
    def post_adapt(self,x):
        pass


class PytorchNumpyImageTransformationAdapter(TransformationAdapter):
    def __init__(self,use_cuda):
        self.use_cuda=use_cuda
    # adapt a pytorch batch of images to process in numpy
    def pre_adapt(self,x):
        x = x.cpu().numpy()
        x= x.transpose((0,2, 3, 1))
        return x

    # adapt a numpy batch of images to process pytorch
    def post_adapt(self,x):
        x = x.transpose(0,3, 1, 2)
        x = torch.from_numpy(x)
        if self.use_cuda:
            x=x.cuda()
        return x




class NumpyPytorchImageTransformationAdapter(TransformationAdapter):
    def __init__(self,use_cuda):
        self.use_cuda=use_cuda
        self.inverse=PytorchNumpyImageTransformationAdapter(use_cuda)

    def pre_adapt(self,x):
        return self.inverse.post_adapt(x)

    def post_adapt(self,x):
        return self.inverse.pre_adapt(x)


from __future__ import annotations

from typing import List
from .base import NumpyMeasure
from .. import MeasureResult
from .activations_iterator import ActivationsIterator
from .stats_running import RunningMeanWelford
import numpy as np

def list_get_all(list:List,indices:List[int])->List:
    return [list[i] for i in indices]

class DistanceFunction:
    def __init__(self,normalize:bool):
        self.normalize=normalize

    def distance(self,batch:np.ndarray,batch_inverted:np.ndarray,mean_running:RunningMeanWelford):
        n_shape=len(batch.shape)
        assert n_shape>=2
        n,f=batch.shape[0],batch.shape[1]

        if n_shape>2 and self.normalize:
            # normalize all extra dimensions
            for i in range(n):
                for j in range(f):
                    batch[i,j,:]/=np.linalg.norm(batch[i,j,:])
                    batch_inverted[i,j,:]/=np.linalg.norm(batch_inverted[i,j,:])

        # ssd of all values
        distances = (batch-batch_inverted)**2
        n_shape=len(batch.shape)
        if n_shape>2:
            # aggregate extra dims to keep only the feature dim
            distances= distances.mean(axis=tuple(range(2,n_shape)))
        distances = np.sqrt(distances)
        assert len(distances.shape)==2
        mean_running.update_all(distances)
    def __repr__(self):
        return f"DF(normalize={self.normalize})"

class DistanceSameEquivarianceSimple(NumpyMeasure):
    def __init__(self, distance_function:DistanceFunction):
        super().__init__()
        self.distance_function=distance_function

    def __repr__(self):
        return f"DSES(df={self.distance_function})"

    def name(self)->str:
        return "Distance Same-Equivariance Simple"
    def abbreviation(self):
        return "DSES"

    def eval(self,activations_iterator:ActivationsIterator,verbose=False)->MeasureResult:
        activations_iterator = activations_iterator.get_both_iterator()
        mean_running=None

        for x, transformation_activations_iterator in activations_iterator.samples_first():
            # transformation_activations_iterator can iterate over all transforms
            for x_transformed, activations,inverted_activations in transformation_activations_iterator:
                if mean_running is None:
                    # do this after the first iteration since we dont know the number
                    # of layers until the first iteration of the activations_iterator
                    mean_running = [RunningMeanWelford() for i in range(len(activations))]
                for j, (layer_activations,inverted_layer_activations) in enumerate(zip(activations,inverted_activations)):
                    self.distance_function.distance(layer_activations,inverted_layer_activations,mean_running[j])
        # calculate the final mean over all samples (and layers)
        means = [b.mean() for b in mean_running]
        return MeasureResult(means,activations_iterator.layer_names(),self)





import numpy as np
import matplotlib.pyplot as plt
import tmeasures as tm
from pathlib import Path
from ..np import AggregateTransformation,AggregateFunction,MeasureTransformation
from .. import MeasureResult
import torch

from tmeasures.np.activations_iterator import ActivationsIterator

def plot_activations(features:np.ndarray, feature_indices:[int], variance_scores:[float], x_transformed:np.ndarray, transformations:tm.TransformationSet, filepath:Path):

    x_transformed= x_transformed.transpose((0, 2, 3, 1))
    n,c = features.shape
    nt,ht,wt,ct = x_transformed.shape

    dpi = int(np.sqrt(n) * 9) + 150
    f, axis = plt.subplots(c+1,n+1,dpi=dpi)
    axis[0, 0].set_title("Inputs")
    axis[0, 0].axis("off")

    for i in range(c):
        axis[i+1, 0].set_title(f"Act {feature_indices[i]} \n, score:{variance_scores[i]:0.2f} ",fontsize=5)
        axis[i+1, 0].axis("off")

    for j in range(n):
        transformed_image= x_transformed[j, :]
        transformed_image-=transformed_image.min()
        transformed_image/= transformed_image.max()
        if ct ==1:
            transformed_image=transformed_image[:,:,0]
        axis[0,j+1].imshow(transformed_image,cmap="gray")
        axis[0,j+1].axis("off")

    for i in range(c):
        mi, ma = features[:,i].min(), features[:,i].max()
        for j in range(n):
            image = features[j, i].reshape((1,1))
            axis[i+1,j+1].imshow(image, vmin=mi, vmax=ma,cmap="gray")
            axis[i+1, j+1].axis("off")

    plt.savefig(filepath)
    plt.close("all")

def plot_feature_maps(feature_maps:np.ndarray, feature_indices:[int], variance_scores:[float], x_transformed:np.ndarray, transformations:tm.TransformationSet, filepath:Path):
    feature_maps = feature_maps.transpose((0,2,3,1))
    x_transformed= x_transformed.transpose((0, 2, 3, 1))
    n,h,w,c = feature_maps.shape
    nt,ht,wt,ct = x_transformed.shape
    if ct == 1:
        x_transformed = x_transformed[:,:, :, 0]

    fontsize=max(3, 10 - int(np.sqrt(n)))
    dpi = int(np.sqrt(n) * 9) + 100
    f, axis = plt.subplots(c+1, n+4, dpi=dpi)
    axis[0, 0].set_title("Inputs")
    axis[0, 0].axis("off")
    for i in range(c):
        title=f"FM {feature_indices[i]}\n, score:\n{variance_scores[i]:0.2f} "
        axis[i+1, 0].text(0,0, title,fontsize=fontsize)
        axis[i+1, 0].axis("off")

    for j in range(n):
        transformed_image  = x_transformed[j, :]
        transformed_image -= transformed_image.min()
        transformed_image /= transformed_image.max()

        axis[0,j+1].imshow(transformed_image)
        axis[0,j+1].axis("off")

    colorbar_images=[]
    for i in range(c):
        mi, ma = feature_maps[:,:,:,i].min(), feature_maps[:,:,:,i].max()
        for j in range(n):
            im=axis[i+1,j+1].imshow(feature_maps[j,:,:,i],vmin=mi,vmax=ma,cmap="gray")
            axis[i+1, j+1].axis("off")
            if j+1 == n:
                colorbar_images.append(im)

    # mean and std of feature maps columns
    for i in range(c):
        mean_feature_map = np.mean(feature_maps[:,:,:,i],axis=0)
        std_feature_map = np.std(feature_maps[:, :, :, i], axis=0)
        axis[i + 1, -3].imshow(mean_feature_map,cmap="gray")
        axis[i + 1, -3].axis("off")
        axis[i + 1, -2].imshow(std_feature_map,cmap="gray")
        axis[i + 1, -2].axis("off")

    for i in range(c):
        axis[i + 1, -1].axis("off")
        cbar = plt.colorbar(colorbar_images[i], ax=axis[i + 1, -1])
        cbar.ax.tick_params(labelsize=fontsize)
    axis[0, -1].axis("off")

        # mean and std of images columns
    axis[0, -3].imshow(x_transformed.mean(axis=0),cmap="gray")
    axis[0, -3].axis("off")

    axis[0, -2].imshow(x_transformed.std(axis=0),cmap="gray")
    axis[0, -2].axis("off")

    plt.savefig(filepath)
    plt.close("all")

def indices_of_smallest_k(a,k):

    indices = np.argpartition(a, k)[:k]
    values = a[indices]
    indices = np.argsort(a)

    # ind.sort()
    return indices[:k]

def indices_of_largest_k(a,k):

    # indices = np.argpartition(a, -k)[:k]
    # values = a[indices]
    indices=np.argsort(a)

    # ind.sort()
    return indices[-k:]

def select_feature_maps(measure_result: tm.MeasureResult, most_invariant_k:int, least_invariant_k:int):
    feature_indices_per_layer=[]
    feature_scores_per_layer = []
    values=measure_result.layers
    layer_names=measure_result.layer_names
    for value,name in zip(values,layer_names):
        if len(value.shape) != 1:
            raise ValueError("Feature maps should be collapsed before calling this function")
        most_invariant_indices = indices_of_smallest_k(value, most_invariant_k)
        least_invariant_indices = indices_of_largest_k(value, least_invariant_k)
        indices = np.concatenate([most_invariant_indices,least_invariant_indices])
        feature_indices_per_layer.append(indices)
        feature_scores_per_layer.append(value[indices])
    return feature_indices_per_layer,feature_scores_per_layer

'''
    Plots the activation of the invariant feature maps determined by :param result
    Plots the best :param features_per_layer feature maps, ie the most invariant
    Creates a plot for each sample image/transformation pair
    '''

from tmeasures.np import MeasureTransformation

def plot_invariant_feature_maps(plot_folderpath:Path, activations_iterator:ActivationsIterator, result:tm.MeasureResult, most_invariant_k:int, least_invariant_k:int
                                #, conv_aggregation:MeasureTransformation
                                ):
    #result=conv_aggregation.apply(result)

    feature_indices_per_layer,invariance_scores_per_layer=select_feature_maps(result, most_invariant_k,least_invariant_k)
    transformations=activations_iterator.get_transformations()

    for i_image,(x,transformation_activations_iterator) in enumerate(activations_iterator.samples_first()):
        layers_activations,x_transformed = activations_iterator.row_from_iterator(transformation_activations_iterator)
        for i_layer,layer_activations in enumerate(layers_activations):
            layer_name=result.layer_names[i_layer]
            filepath = plot_folderpath / f"image{i_image}_layer{i_layer}_{layer_name}.png"
            feature_indices=feature_indices_per_layer[i_layer]
            invariance_scores=invariance_scores_per_layer[i_layer]
            #only plot conv layer activations
            if len(layer_activations.shape)==4:
                feature_maps=layer_activations[:,feature_indices,:,:]
                plot_feature_maps(feature_maps,feature_indices,invariance_scores,x_transformed,transformations,filepath)
            elif len(layer_activations.shape)==2:
                features = layer_activations[:, feature_indices]
                plot_activations(features, feature_indices, invariance_scores, x_transformed, transformations,filepath)







import tmeasures as tm


from tmeasures import MeasureResult
from .activations_iterator import ActivationsIterator
from .stats_running import RunningMeanAndVarianceWelford, RunningMeanWelford,RunningMeanVarianceSets
from .base import NumpyMeasure
from .quotient import divide_activations
from tqdm import tqdm
from .layer_transformation import MeasureTransformation,IdentityTransformation
import numpy as np

class TransformationVarianceInvariance(NumpyMeasure):
    def __init__(self):
        super().__init__()

    def __repr__(self):
        return f"{self.abbreviation()}"

    def eval(self, activations_iterator: ActivationsIterator,verbose=False) -> MeasureResult:
        layer_names = activations_iterator.layer_names()
        n_layers = len(layer_names)
        mean_running = [RunningMeanWelford() for i in range(n_layers)]

        for x,transformation_activations  in tqdm(activations_iterator.samples_first(),disable=not verbose):

            #calculate the running mean/variance/std over all transformations of x
            transformation_variances_running = [RunningMeanVarianceSets() for i in range(n_layers)]
            # print(transformation_variances_running[0])
            for x_transformed, activations in transformation_activations:
                for i, layer_activations in enumerate(activations):
                    # update the mean over all transformations for this sample
                    transformation_variances_running[i].update_batch(layer_activations.astype(np.double))

            # update the mean with the numpy sample of all transformations of x
            for i in range(n_layers):
                layer_measure = transformation_variances_running[i].std()

                # print(layer_measure)
                mean_running[i].update(layer_measure)

        # calculate the final mean over all samples (for each layer)
        mean_variances = [b.mean() for b in mean_running]
        return MeasureResult(mean_variances, layer_names, self)

class SampleVarianceInvariance(NumpyMeasure):
    def __init__(self):
        super().__init__()

    def __repr__(self):
        return f"{self.abbreviation()}"


    def eval(self, activations_iterator: ActivationsIterator,verbose=False) -> MeasureResult:
        layer_names = activations_iterator.layer_names()
        n_layers = len(layer_names)
        mean_variances_running = [RunningMeanWelford() for i in range(n_layers)]

        for transformation, samples_activations_iterator in tqdm(activations_iterator.transformations_first(),disable=not verbose):
            samples_variances_running = [RunningMeanAndVarianceWelford() for i in range(n_layers)]

            # calculate the variance of all samples for this transformation
            for x, batch_activations in samples_activations_iterator:
                for j, layer_activations in enumerate(batch_activations):
                    samples_variances_running[j].update_all(layer_activations)
            # update the mean over all transformation (and layers)
            for layer_mean_variances_running, layer_samples_variance_running in zip(mean_variances_running,samples_variances_running):
                samples_variance = layer_samples_variance_running.std()
                layer_mean_variances_running.update(samples_variance)

        # calculate the final mean over all transformations (and layers)

        mean_variances = [b.mean() for b in mean_variances_running]
        return MeasureResult(mean_variances, layer_names, self)


class NormalizedVarianceInvariance(NumpyMeasure):
    def __init__(self, pre_normalization_transformation: MeasureTransformation=IdentityTransformation()):
        self.sv = SampleVarianceInvariance()
        self.tv = TransformationVarianceInvariance()
        self.pre_normalization_transformation = pre_normalization_transformation

    def eval(self, activations_iterator: ActivationsIterator,verbose=False) -> MeasureResult:
        tv_result=self.tv.eval(activations_iterator,verbose)
        sv_result=self.sv.eval(activations_iterator,verbose)

        tv_result = self.pre_normalization_transformation.apply(tv_result)
        sv_result = self.pre_normalization_transformation.apply(sv_result)

        result=divide_activations(tv_result.layers, sv_result.layers)
        return MeasureResult(result, activations_iterator.layer_names(), self)

    def __repr__(self):
        ca = f"pnt={self.pre_normalization_transformation}"
        return f"{self.abbreviation()}({ca})"



import abc
from typing import Iterable, List

import torch

from tmeasures import Transformation, TransformationSet


class PyTorchTransformation(Transformation,torch.nn.Module):

    @abc.abstractmethod
    def parameters(self)->torch.Tensor:
        pass


class PyTorchTransformationSet(TransformationSet,Iterable[PyTorchTransformation]):

    def parameter_range(self)->List[torch.Tensor]:
        parameters = torch.stack([t.parameters() for t in self],dim=0)
        mi,_ = parameters.min(dim=0)
        ma,_ = parameters.max(dim=0)
        return mi, ma


class IdentityTransformation(PyTorchTransformation):
    
    def parameters(self) -> torch.Tensor:
        return 0
     
    def __call__(self, x):
        return x


class IdentityTransformationSet(PyTorchTransformationSet):
    
    def __init__(self):
        super().__init__([IdentityTransformation()])


from __future__ import annotations

from typing import Tuple
import torch
from tmeasures import TransformationSet
import tmeasures as tm

def list_get_all(list:list,indices:list[int])->[]:
    return [list[i] for i in indices]


class ActivationsTransformer:
    def __init__(self,activation_shapes: [Tuple[int,]], layer_names: [str],
                         transformation_set: tm.TransformationSet,inverse:bool):

        self.layer_names,self.indices=self.get_valid_layers(activation_shapes,layer_names,transformation_set)


        self.valid_shapes=list_get_all(activation_shapes,self.indices)

        self.transformation_sets=self.get_transformations_set(self.valid_shapes, transformation_set, inverse)
        assert len(self.transformation_sets) == len(self.valid_shapes)
        m= len(transformation_set)
        for s in self.transformation_sets:
            mi=len(s)
            assert mi == m, f"Transformation sets for each layer should have the same number of transformations ({mi}) as the original transformation set ({m})"



    def get_valid_layers(self, activation_shapes: [Tuple[int,]], layer_names: [str],
                         transformation_set: tm.TransformationSet)->([str],[int]):
        # get indices of layers for which the transformation is valid
        indices = [i for i, shape in enumerate(activation_shapes) if transformation_set.valid_input(shape)]
        # keep only this layers
        layer_names = list_get_all(layer_names, indices)

        return layer_names, indices

    def get_transformations_set(self, shapes: [Tuple[int,]],
                                transformation_set: tm.TransformationSet,inverse:bool):
        transformation_sets = []

        for s in shapes:
            n, c, h, w = s
            layer_transformation_set: tm.TransformationSet = transformation_set.copy()
            # print(len(transformation_set),"vs",len(layer_transformation_set))
            # layer_transformation_set.set_pytorch(False)
            #layer_transformation_set.set_input_shape((h, w, c))
            layer_transformation_set_list = list(layer_transformation_set)
            if inverse:
                layer_transformation_set_list = [l.inverse() for l in layer_transformation_set_list]
            transformation_sets.append(layer_transformation_set_list)
        return transformation_sets

    def filter_activations(self,activations: [torch.Tensor])->[torch.Tensor]:
        return list_get_all(activations,self.indices)

    def trasform_st_same_row(self, activations: [torch.Tensor],
                             t_start: int, t_end: int):
        # iterate over each layer and corresponding layer transformations
        for layer_activations, layer_transformations in zip(activations, self.transformation_sets):
            # each sample of the layer activations corresponds to a different column of the st matrix
            # => a different transformation
            # t_start and t_end indicate the corresponding column indices
            for i, transformation in enumerate(layer_transformations[t_start:t_end]):
                transformed_activations = transformation(layer_activations[i,:])
                # print(fm.shape, inverse_fm.shape)
                layer_activations[i,:] = transformed_activations

    def trasform_st_same_column(self, activations: [torch.Tensor], t_i: int):
        for layer_activations, layer_transformations in zip(activations, self.transformation_sets):
            # each sample of the layer activations corresponds to a different row of the st matrix
            # => a different sample
            # t_i indicate the corresponding column index, that is, the transformation index
            transformation = layer_transformations[t_i]
            for i in range(layer_activations.shape[0]):
                layer_activations[i,:] = transformation(layer_activations[i,:])






from .activations_iterator import ActivationsIterator
import numpy as np
from typing import  List
from .. import MeasureResult, StratifiedMeasureResult
from ..measure import Measure

import abc


class NumpyMeasure(Measure):

    @abc.abstractmethod
    def eval(self,activations_iterator:ActivationsIterator,verbose=False)-> MeasureResult:
        '''

        '''
        pass


    def eval_stratified(self,classes_iterators:List[ActivationsIterator],class_labels:List[str])-> StratifiedMeasureResult:
        '''
        Calculate the `variance_measure` for each class separately
        Also calculate the average stratified `variance_measure` over all classes
        '''
        variance_per_class = [self.eval(iterator) for iterator in classes_iterators]
        stratified_measure_layers = self.mean_variance_over_classes(variance_per_class)

        layer_names= variance_per_class[0].layer_names

        return StratifiedMeasureResult(stratified_measure_layers, layer_names, self, variance_per_class, class_labels)


    def mean_variance_over_classes(self, class_variance_result:List[MeasureResult]) -> List[np.ndarray]:
        # calculate the mean activation of each unit in each layer over the set of classes
        class_variance_layers=[r.layers for r in class_variance_result]
        # class_variance_layers is a list (classes) of list layers)
        # turn it into a list (layers) of lists (classes)
        layer_class_vars=[list(i) for i in zip(*class_variance_layers)]
        # compute average variance of each layer over classses
        layer_vars=[ sum(layer_values)/len(layer_values) for layer_values in layer_class_vars]
        return layer_vars




from __future__ import annotations

from torch import  nn
import torch
import abc
from typing import Callable, List,MutableMapping,Any, Tuple,Union

from tmeasures.utils import get_all


class ActivationsModule(nn.Module):

    @abc.abstractmethod
    def activation_names(self) -> List[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def forward_activations(self, args) -> List[torch.Tensor]:
        raise NotImplementedError()

    def n_activations(self):
        return len(self.activation_names())
#


ActivationFilter = Callable[[ActivationsModule,str],bool]
class FilteredActivationsModule(ActivationsModule):
    def __init__(self, inner_model:ActivationsModule, activations_filter:ActivationFilter):
        super().__init__()
        self.inner_model = inner_model
        original_names = inner_model.activation_names()
        self.names = [name for name in original_names if activations_filter(inner_model,name)]
        assert len(self.names)>0
        self.indices = [original_names.index(n) for n in self.names]
    def forward(self):
        return self.inner_model.forward()
    def activation_names(self) -> List[str]:
        return self.names
    def eval(self):
        self.inner_model.eval()
    @abc.abstractmethod
    def forward_activations(self, args) ->  List[torch.Tensor]:
        activations = self.inner_model.forward_activations(args)
        activations = [activations[i] for i in self.indices]
        return activations



def intersect_lists(a:List,b:List):
    return list(set(a) & set(b))

class DuplicateKeyError(ValueError):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

Input = Union[MutableMapping,Any]
def flatten_dict(d_or_val:Input, prefix='', sep='/', allow_repeated=False)->dict[str,Any]:
    print(d_or_val)
    if isinstance(d_or_val, MutableMapping):
        result = {}
        if prefix =="":
            prefix_sep = ""
        else:
            prefix_sep = prefix+sep
        for k, v in d_or_val.items():
            new_prefix = prefix_sep+str(k)
            flattened_child = flatten_dict(v,new_prefix,sep)
            if not allow_repeated :
                intersection = intersect_lists(flattened_child.keys(),d_or_val.keys())
                if len(intersection)>0:
                    raise DuplicateKeyError(intersection)
            result.update(flattened_child)
        return result
    else :
        return { prefix : d_or_val }

def flatten_dict_list(d_or_val:Input,key="",full_name=True)->List[Tuple[str,Any]]:
    if isinstance(d_or_val, MutableMapping):
        result = []
        for k, v in d_or_val.items():
            if full_name:
                new_key = key + "/" + k
            else:
                new_key = k
            flattened_child = flatten_dict_list(v,key=new_key)
            result+=flattened_child
        return result
    else :
        return [(key, d_or_val)]


def named_children_deep(m: torch.nn.Module):
    children = dict(m.named_children())
    if children == {}:
        return m
    else:
        output = {}
        for name, child in children.items():
            if name.isnumeric():
                key = child.__class__.__name__+"_"+name
            else:
                key=name
            if key in output:
                raise DuplicateKeyError
            try:
                output[key] = named_children_deep(child)
            except TypeError:
                output[key] = named_children_deep(child)
        return output

class AutoActivationsModule(ActivationsModule):
    def __init__(self,module:nn.Module,full_name=True,filter:Callable=lambda x: True) -> None:
        super().__init__()
        self.reset_values()
        self.module = module
        self.children_tree = named_children_deep(module)
        
        self.names,self.activations = zip(*flatten_dict_list(self.children_tree,full_name=full_name))
        filter_indices = [i for i,a in enumerate(self.activations) if filter(a)]
        self.names=get_all(self.names,filter_indices)
        self.activations=get_all(self.activations,filter_indices)

        self.register_hooks(self.activations)
        

    def register_hooks(self,activations:List[nn.Module]):
        def store_activation(index):
            def hook(model, input, output):
                self.values[index] = output.detach()
            return hook

        for i,v in enumerate(activations):
            v.register_forward_hook(store_activation(i))

    def reset_values(self):
        self.values = {} 

    def forward_activations(self, args) -> List[torch.Tensor]:
        '''
        This function is not thread safe.
        '''
        # clear values
        self.reset_values()
        # call model, hooks are executed
        self.module.forward(args)
        
        # Check that all original activations are present in the values
        nv,na = len(self.values),len(self.activations)
        if nv!=na:
            indices_a = set(list(range(len(self.activations))))
            indices_v = set(list(self.values.keys()))
            diff = indices_a-indices_v
            missing_names = [self.activations[i] for i in diff]
            assert nv==na, f"Values output by network ({nv}) dont match original number of activations ({na}). Values: \n {missing_names}"
        # transform to list and ensure order of values is same as order of activation names
        activations_list = [self.values[i] for i in range(len(self.activations))]
        return activations_list

    def activation_names(self) -> List[str]:
        return self.names


from tmeasures.pytorch.base import PyTorchLayerMeasure, STMatrixIterator
from tmeasures.pytorch.stats_running import RunningMeanWelford, RunningMeanAndVarianceWelford

class Variance(PyTorchLayerMeasure):
    def eval(self, st_iterator: STMatrixIterator, layer_name: str):
        mean = RunningMeanWelford()
        
        for row, row_iterator in enumerate(st_iterator):
            # print(f"frow {layer_name}",flush=True)
            row_variance = RunningMeanAndVarianceWelford()
            for batch_n, batch_activations in enumerate(row_iterator):
                # print(f"fcol {layer_name}",flush=True)
                row_variance.update_batch(batch_activations.double())
            row_std = row_variance.std()
            mean.update(row_std)
        return mean.mean()

# class Distance(PyTorchLayerMeasure):
#     def eval(self, st_iterator:STMatrixIterator):
#         mean = RunningMeanWelford()
#         for row,row_iterator in enumerate(st_iterator):
#             row_variance = RunningMeanAndVarianceWelford()
#             for batch_n,batch_activations in enumerate(row_iterator):
#                 row_variance.update_all(batch_activations.double())
#             row_std = row_variance.std()
#             mean.update(row_std)
#         return mean.mean()
#



from tmeasures.np.stats_running import RunningMeanAndVarianceWelford, RunningMeanWelford,RunningMean,RunningVariance,RunningMeanVarianceSets,RunningMeanSets
import pytest

import numpy as np
import itertools
eps = 1e-5
atol = 0.1
dimensions=[10]
n_samples = [100,1000]
scales = [1e-5, 1, 100]
centers = [-5, 0, 10]
batch_size = 10
for n in n_samples:
    assert n % batch_size == 0

values = [np.random.randn(n,dim)*s+c for (n,dim,s,c) in itertools.product(n_samples,dimensions,scales,centers)]

values_batched =[np.random.randn(n//batch_size,batch_size,dim)*s+c for (n,dim,s,c) in itertools.product(n_samples,dimensions,scales,centers)]

values = [v.astype(np.double) for v in values]
values_batched = [v.astype(np.double) for v in values_batched]

def compare_variance_batched(m:RunningVariance,x:np.ndarray):

    for i in range(x.shape[0]):
        m.update_batch(x[i,:, :])
    expected = x.var(axis=(0,1),ddof=1)
    actual = m.var()
    assert np.sum(actual < 0) == 0, "Variance can't be negative"
    # assert expected == pytest.approx(actual,rel=eps,abs=atol)
    np.testing.assert_allclose(expected, actual, rtol=eps,atol=atol)

def compare_variance(m:RunningVariance,x:np.ndarray):
    
    for i in range(x.shape[0]):
        m.update(x[i, :])

    expected = x.var(axis=0,ddof=1)
    actual = m.var()
    assert np.sum(actual < 0) == 0, "Variance can't be negative"
    np.testing.assert_allclose(expected, actual, rtol=eps)


def compare_means_batched(m:RunningMean,x:np.ndarray):
    for i in range(x.shape[0]):
        m.update_batch(x[i,:, :])
    expected = x.mean(axis=(0,1))
    actual = m.mean()
    np.testing.assert_allclose(expected, actual, rtol=eps)

def compare_means(m:RunningMean,x:np.ndarray):
    for i in range(x.shape[0]):
        m.update(x[i, :])
    expected = x.mean(axis=0)
    actual = m.mean()
    np.testing.assert_allclose(expected, actual, rtol=eps,atol=atol)

@pytest.mark.parametrize("x",values)
def test_mean_welford(x:np.ndarray):
    compare_means(RunningMeanWelford(),x)

@pytest.mark.parametrize("x",values_batched)
def test_mean_welford(x:np.ndarray):
    compare_means_batched(RunningMeanWelford(),x)

@pytest.mark.parametrize("x",values)
def test_mean_sets(x:np.ndarray):
    compare_means(RunningMeanSets(),x)

@pytest.mark.parametrize("x",values_batched)
def test_mean_sets_batched(x:np.ndarray):
    compare_means_batched(RunningMeanSets(),x)

@pytest.mark.parametrize("x",values)
def test_mean_variance_welford(x:np.ndarray):
    compare_means(RunningMeanAndVarianceWelford(),x)

@pytest.mark.parametrize("x",values_batched)
def test_mean_variance_welford(x:np.ndarray):
    compare_means_batched(RunningMeanAndVarianceWelford(),x)

@pytest.mark.parametrize("x",values)
def test_variance_sets(x:np.ndarray):
    compare_variance(RunningMeanVarianceSets(),x)

@pytest.mark.parametrize("x",values_batched)
def test_variance_sets_batched(x:np.ndarray):
    compare_variance_batched(RunningMeanVarianceSets(),x)

@pytest.mark.parametrize("x",values)
def test_variance_welford(x:np.ndarray):
    compare_variance(RunningMeanAndVarianceWelford(),x)

@pytest.mark.parametrize("x",values_batched)
def test_variance_welford(x:np.ndarray):
    compare_variance_batched(RunningMeanAndVarianceWelford(),x)



from tmeasures.pytorch.stats_running import RunningMeanAndVarianceWelford, RunningMeanWelford,RunningMean,RunningVariance,RunningMeanSets,RunningMeanVarianceSets
import itertools
import torch
import pytest

eps = 1e-5
abs = 1e-2
dimensions=[10]
n_samples = [100,1000]
scales = [ 1e-5, 1, 10]
centers = [-5, 0, 10]
batch_size = 10
for n in n_samples:
    assert n % batch_size == 0

values = [torch.randn(n,dim)*s+c for (n,dim,s,c) in itertools.product(n_samples,dimensions,scales,centers)]
values_batched =[torch.randn(n//batch_size,batch_size,dim)*s+c for (n,dim,s,c) in itertools.product(n_samples,dimensions,scales,centers)]

# use doubles
values = [v.double() for v in values]
values_batched = [v.double() for v in values_batched]

def compare_values(expected:torch.Tensor,actual:torch.Tensor):
    print("Expected:")
    print(expected)
    print("Actual:")
    print(actual)

    return torch.all(torch.isclose(expected,actual,rtol=eps))


def compare_variance_batched(m:RunningVariance,x:torch.Tensor):
    for i in range(x.shape[0]):
        m.update_batch(x[i,:, :])
    expected = x.var(dim=(0,1))
    actual = m.var()
    assert torch.sum(actual < 0) == 0, f"Variance can't be negative {actual}"
    torch.testing.assert_close(actual,expected,rtol=eps,atol=abs)


def compare_variance(m:RunningVariance,x:torch.Tensor):
    for i in range(x.shape[0]):
        m.update(x[i, :])
    expected = x.var(dim=0)
    actual = m.var()
    assert torch.sum(actual < 0) == 0, f"Variance can't be negative {actual}"
    torch.testing.assert_close(actual, expected, rtol=eps, atol=abs)


def compare_means(m:RunningMean,x:torch.Tensor):
    for i in range(x.shape[0]):
        m.update(x[i, :])
    expected = x.mean(dim=0)
    actual = m.mean()
    torch.testing.assert_close(actual, expected, rtol=eps, atol=abs)


def compare_means_batched(m:RunningMean,x:torch.Tensor):
    for i in range(x.shape[0]):
        m.update_batch(x[i,:, :])
    expected = x.mean(dim=(0,1))
    actual = m.mean()
    torch.testing.assert_close(actual, expected, rtol=eps, atol=abs)


@pytest.mark.parametrize("x", values)
def test_mean_welford(x: torch.Tensor):
    compare_means(RunningMeanWelford(), x)


@pytest.mark.parametrize("x", values_batched)
def test_mean_welford(x: torch.Tensor):
    compare_means_batched(RunningMeanWelford(), x)


@pytest.mark.parametrize("x", values)
def test_mean_sets(x: torch.Tensor):
    compare_means(RunningMeanSets(), x)


@pytest.mark.parametrize("x", values_batched)
def test_mean_sets_batched(x: torch.Tensor):
    compare_means_batched(RunningMeanSets(), x)


@pytest.mark.parametrize("x", values)
def test_mean_variance_welford(x: torch.Tensor):
    compare_means(RunningMeanAndVarianceWelford(), x)


@pytest.mark.parametrize("x", values_batched)
def test_mean_variance_welford(x: torch.Tensor):
    compare_means_batched(RunningMeanAndVarianceWelford(), x)


@pytest.mark.parametrize("x", values)
def test_variance_sets(x: torch.Tensor):
    compare_variance(RunningMeanVarianceSets(), x)


@pytest.mark.parametrize("x", values_batched)
def test_variance_sets_batched(x: torch.Tensor):
    print(f"Shape of input:{x.shape}, (μ={x.mean(dim=(0,1))},σ={x.var(dim=(0,1))}\n")
    compare_variance_batched(RunningMeanVarianceSets(), x)

@pytest.mark.parametrize("x",values)
def test_variance_welford(x:torch.Tensor):
    compare_variance(RunningMeanAndVarianceWelford(),x)

@pytest.mark.parametrize("x",values_batched)
def test_variance_welford(x:torch.Tensor):
    compare_variance_batched(RunningMeanAndVarianceWelford(),x)



from __future__ import annotations
from typing import List, Tuple
import numpy as np
import abc
import re
from .utils import get_all

ActivationsByLayer = List[np.ndarray]

# TODO change `layer` for `activation` in variable/methods to unify vocabulary

class MeasureResult:
    def __init__(self,layers:ActivationsByLayer,layer_names:List[str],measure:Measure,extra_values=dict()):
        assert (len(layers) == len(layer_names)), f"The number of layers {len(layers)} should match the length of the layer_names list {layer_names} "
        self.layers=layers
        self.layer_names=layer_names
        self.measure=measure
        self.extra_values=extra_values

    def __repr__(self):
        return f"MeasureResult({self.measure})"

    def all_1d(self):
        return np.any([ len(l.shape)==1 for l in self.layers])

    def per_layer_average(self) -> np.ndarray:
        result = []
        for layer in self.layers:
            layer=layer[:]
            layer_average=layer[np.isfinite(layer)].mean()
            result.append(layer_average)
        return np.array(result)

    def per_layer_mean_std(self) -> Tuple[np.ndarray,np.ndarray]:
        means = []
        stds = []
        for layer in self.layers:
            layer=layer[:]
            layer_mean=layer[np.isfinite(layer)].mean()
            layer_std=layer[np.isfinite(layer)].std()
            means.append(layer_mean)
            stds.append(layer_std)
        return np.array(means),np.array(stds)

    def remove_layers(self,remove_indices:List[int])->MeasureResult:
        n = len(self.layer_names)
        all_indices=set(list(range(n)))
        keep_indices = list(all_indices.difference(set(remove_indices)))
        layers = get_all(self.layers,keep_indices)
        layer_names = get_all(self.layer_names,keep_indices)
        return MeasureResult(layers,layer_names,self.measure,self.extra_values)

    def weighted_global_average(self):
        return self.per_layer_average().mean()

    def global_average(self)-> float:
        return self.per_layer_average().mean()

    def layers_dict(self):
        return dict(zip(self.layer_names,self.layers))



class StratifiedMeasureResult(MeasureResult):
    def __init__(self, layers:ActivationsByLayer, layer_names:List[str], measure:Measure,
                 results:List[MeasureResult], labels:List[str]):
        super().__init__(layers,layer_names,measure)
        self.results=results
        self.labels=labels

    def __repr__(self):
        return f"StratifiedMeasureResult {self.measure}"




class Measure(abc.ABC):
    def __repr__(self):
        return f"{self.__class__.__name__}"

    def id(self):
        return str(self)

    # @abc.abstractmethod
    # def name(self):
    #     pass
    #
    # @abc.abstractmethod
    # def abbreviation(self):
    #     pass

    # Returns the name of the measure, separated by spaces
    def name(self):
        # get first part before parenthesis (if any)
        class_name = self.__class__.__name__
        result = class_name.split("(")[0]
        # add a space before capitalized words
        result = re.sub( r"([A-Z])", r" \1", result)
        return result

    # generates an abbreviation of the measure, based on its name
    def abbreviation(self):
        initials=re.findall('[A-Z]', self.name())
        return "".join(initials)

    @abc.abstractmethod
    def eval(self)->MeasureResult:
        pass


import torch
import abc

from . import PyTorchMeasureResult


class MeasureTransformation(torch.nn.Module):

    @abc.abstractmethod
    def forward(self, x: torch.Tensor, activation_name: str) -> torch.Tensor:
        pass

    def transform_result(self, r: PyTorchMeasureResult):
        for i, (l, n) in enumerate(zip(r.layers, r.layer_names)):
            r.layers[i] = self.forward(l, n)

    def __repr__(self):
        return self.__class__.__name__


class NoTransformation(MeasureTransformation):

    def forward(self, x: torch.Tensor, activation_name: str) -> torch.Tensor:
        return x


class AverageExtraDimensions(MeasureTransformation):

    def forward(self, x: torch.Tensor, activation_name: str) -> torch.Tensor:
        dims = tuple(range(1, len(x.shape)))
        return x.mean(dim=dims)


class AverageFeatureMaps(MeasureTransformation):
    def __init__(self, channel_dimension: int = 0):
        super().__init__()
        self.channel_dimension = channel_dimension

    def forward(self, x: torch.Tensor, activation_name: str) -> torch.Tensor:
        d = len(x.shape)
        if d != 3:
            # ignore non-feature map activations
            return x

        dims = tuple(filter(lambda n: n != self.channel_dimension, range(d)))

        return x.mean(dim=dims)

    def __repr__(self):
        extra = ""
        if self.channel_dimension!=0:
            extra = f"(dim={self.channel_dimension})"
        return f"{super().__repr__()}{extra}"



from queue import Queue, Empty
from time import sleep
from typing import Sized, Iterable


class IterableQueue(Sized, Iterable):
    """Queue supporting Iterator and Sized protocols.
    Queue has a max size so it can be iterated upon with a for loop
    """

    def __init__(self, n: int, maxsize=None,name=""):
        if maxsize is None:
            maxsize = n
        self.queue = Queue(maxsize=maxsize)
        self.name=name
        self.n = n
        self.i = 0
        self.stop=False

    def __len__(self):
        return self.n

    def put(self, x,block=True,timeout=None):
        self.queue.put(x,block=block,timeout=timeout)

    def __iter__(self):
        self.i = 0
        return self
    def get(self,block=True,timeout=None):
        return self.queue.get(block=block,timeout=timeout)
    
    def __next__(self):
        # print(f"next {self.i}/{self.n} (maxsize {self.maxsize})")
        self.i += 1
        if self.i == self.n + 1 or self.stop:
            raise StopIteration()
        # print(f"ITQUEUE {self.name}: Getting item...")
        # item = None
        # while item is None:
        #     try:
        #         item = self.queue.get_nowait()
        #     except Empty:
        #         # print(f"{self.name} sleeping")
        #         sleep(0)
        # print(f"ITQUEUE {self.name}: obtained item: {item}...")
        item = self.queue.get()
        return item



from tmeasures.np.activations_iterator import ActivationsIterator
from .model import ActivationsModule
import torch
from torch.utils.data import DataLoader,Dataset

from tmeasures import TransformationSet, Transformation
from tmeasures.adapters import TransformationAdapter


import tmeasures as tm
import abc
from .activations_transformer import ActivationsTransformer




class PytorchActivationsIterator(ActivationsIterator):

    def __init__(self, model: ActivationsModule, dataset:Dataset, transformations: TransformationSet, batch_size,
                 num_workers, use_cuda,adapter: TransformationAdapter=None):

        self.model = model
        self.dataset = dataset
        self.transformations = transformations

        self.dataset = dataset
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.use_cuda = use_cuda
        self.adapter = adapter

    @abc.abstractmethod
    def samples_activation(self, t_i: int, transformation: Transformation, dataloader: DataLoader):
        pass

    @abc.abstractmethod
    def transformations_activations(self, x: torch.Tensor):
        pass

    def get_transformations(self) ->tm.TransformationSet:
        return self.transformations
    def transformations_first(self):
        for t_i, transformation in enumerate(self.transformations):
            dataloader = DataLoader(self.dataset, batch_size=self.batch_size, shuffle=False,num_workers=self.num_workers, drop_last=False, pin_memory=True)
            yield transformation, self.samples_activation(t_i, transformation, dataloader)

    '''
         Returns the activations of the models by iterating first over transformations and 
         then, for each transformation, over samples
     '''

    def samples_first(self):
        dataloader = DataLoader(self.dataset, batch_size=self.batch_size, shuffle=False,
                                num_workers=self.num_workers, drop_last=False)

        with torch.no_grad():
            for batch in dataloader:
                batch_cpu = batch
                if self.use_cuda:
                    batch = batch.cuda()
                for i in range(batch.shape[0]):
                    x = batch[i, :]
                    yield batch_cpu[i, :], self.transformations_activations(x)

    def transform_sample(self, x: torch.Tensor):
        x = x.unsqueeze(0)
        results = []
        for i, transformation in enumerate(self.transformations):
            transformed = self.transform_batch(transformation, x)
            results.append(transformed)
        result = torch.cat(results)
        return result

    def transform_batch(self, transformation, x: torch.Tensor):
        if not self.adapter is None:
            x = self.adapter.pre_adapt(x)
        results = []
        for i in range(x.shape[0]):
            results.append(transformation(x[i,:]))
        x = torch.stack(tuple(results),dim=0)

        if not self.adapter is None:
            x = self.adapter.post_adapt(x)
        return x

    def layer_names(self):
        return self.model.activation_names()

    def get_inverted_activations_iterator(self) -> ActivationsIterator:
        return InvertedPytorchActivationsIterator(self.model, self.dataset, self.transformations, self.batch_size,
                                                  self.num_workers,  self.use_cuda,self.adapter)

    def get_both_iterator(self) -> ActivationsIterator:
        return BothPytorchActivationsIterator(self.model, self.dataset, self.transformations, self.batch_size,
                                              self.num_workers,  self.use_cuda,self.adapter,)

    def get_normal_activations_iterator(self) -> ActivationsIterator:
        return NormalPytorchActivationsIterator(self.model, self.dataset, self.transformations, self.batch_size,
                                                self.num_workers, self.use_cuda,self.adapter,)


class NormalPytorchActivationsIterator(PytorchActivationsIterator):

    def samples_activation(self, t_i, transformation, dataloader):
        for batch in dataloader:
            if self.use_cuda:
                batch = batch.cuda()
            batch = self.transform_batch(transformation, batch)
            with torch.no_grad():
                batch_activations = self.model.forward_activations(batch)
                batch_activations = [a.cpu().numpy() for a in batch_activations]
            yield batch, batch_activations

    def transformations_activations(self, x):

        x_transformed = self.transform_sample(x)
        dataloader = DataLoader(x_transformed, batch_size=self.batch_size, shuffle=False,
                                num_workers=0, drop_last=False)
        for batch in dataloader:
            with torch.no_grad():
                batch_activations = self.model.forward_activations(batch)
            batch_activations = [a.cpu().numpy() for a in batch_activations]
            yield batch, batch_activations


class InvertedPytorchActivationsIterator(PytorchActivationsIterator):

    def samples_activation(self, t_i, transformation, dataloader):
        self.activations_transformer = None
        for batch in dataloader:
            if self.use_cuda:
                batch = batch.cuda()
            batch = self.transform_batch(transformation, batch)
            with torch.no_grad():
                batch_activations = self.model.forward_activations(batch)
                if self.activations_transformer is None:
                    shapes = [a.shape for a in batch_activations]
                    self.activations_transformer = ActivationsTransformer(shapes, self.model.activation_names(),
                                                                          self.get_transformations(), inverse=True)
                # filter activations to those accepted by the transformations
                batch_activations = self.activations_transformer.filter_activations(batch_activations)

                # inverse transform selected activations
                self.activations_transformer.trasform_st_same_column(batch_activations, t_i)
                batch_activations = [a.cpu().numpy() for a in batch_activations]
                yield batch, batch_activations

    '''
         Returns the activations of the models by iterating first over transformations and 
         then, for each transformation, over samples
     '''

    def transformations_activations(self, x):
        x_transformed = self.transform_sample(x)
        dataloader = DataLoader(x_transformed, batch_size=self.batch_size, shuffle=False,
                                num_workers=0, drop_last=False)
        self.activations_transformer = None
        t_start = 0
        for batch in dataloader:
            with torch.no_grad():
                batch_activations = self.model.forward_activations(batch)
            t_end = t_start + batch.shape[0]
            if self.activations_transformer is None:
                shapes = [a.shape for a in batch_activations]
                self.activations_transformer = ActivationsTransformer(shapes, self.model.activation_names(), self.transformations, inverse=True)
            # filter activations to those accepted by the transformations
            batch_activations = self.activations_transformer.filter_activations(batch_activations)
            # inverse transform selected activations
            self.activations_transformer.trasform_st_same_row(batch_activations, t_start, t_end)
            batch_activations = [a.cpu().numpy() for a in batch_activations]
            t_start = t_end
            yield batch, batch_activations

    def get_inverted_activations_iterator(self) -> ActivationsIterator:
        return self

    def layer_names(self) -> [str]:
        return self.activations_transformer.layer_names


class BothPytorchActivationsIterator(PytorchActivationsIterator):

    def samples_activation(self, t_i, transformation, dataloader):
        for batch in dataloader:
            if self.use_cuda:
                batch = batch.cuda()

            transformed_batch = self.transform_batch(transformation, batch)
            with torch.no_grad():
                pre_transformed_activations = self.model.forward_activations(transformed_batch)
                post_transformed_activations = self.model.forward_activations(batch)
            if self.activations_transformer is None:
                shapes = [a.shape for a in pre_transformed_activations]
                self.activations_transformer = ActivationsTransformer(shapes, self.model.activation_names(),
                                                                      self.get_transformations(),False)
            # filter activations to those accepted by the transformations
            pre_transformed_activations = self.activations_transformer.filter_activations(pre_transformed_activations)
            post_transformed_activations = self.activations_transformer.filter_activations(post_transformed_activations)
            # print([a.shape for a in post_transformed_activations])
            # transform selected activations
            self.activations_transformer.trasform_st_same_column(post_transformed_activations, t_i)
            # print([a.shape for a in post_transformed_activations])
            pre_transformed_activations = [a.cpu().numpy() for a in pre_transformed_activations]
            post_transformed_activations = [a.cpu().numpy() for a in post_transformed_activations]
            yield batch, pre_transformed_activations, post_transformed_activations

    '''
         Returns the activations of the models by iterating first over transformations and 
         then, for each transformation, over samples
     '''

    def transformations_activations(self, x):
        x_transformed = self.transform_sample(x)
        dataloader = DataLoader(x_transformed, batch_size=self.batch_size, shuffle=False,
                                num_workers=0, drop_last=False)
        t_start = 0
        # calculate activations of untransformed sample
        with torch.no_grad():
            untransformed_activations = self.model.forward_activations(x.unsqueeze(0))
        # Generate activations transformers
        shapes = [a.shape for a in untransformed_activations]
        self.activations_transformer = ActivationsTransformer(shapes, self.model.activation_names(), self.transformations,False)

        # remove activations that can't be transformed
        untransformed_activations = self.activations_transformer.filter_activations(untransformed_activations)

        for batch in dataloader:
            with torch.no_grad():
                pre_transformed_activations = self.model.forward_activations(batch)

            b_n=batch.shape[0]
            t_end = t_start + b_n
            # filter activations to those accepted by the transformations
            pre_transformed_activations = self.activations_transformer.filter_activations(pre_transformed_activations)
            # calculate post transformed activations
            # post_transformed_activations = [a.clone() for a in untransformed_activations]
            # replicate to batch size b_n
            post_transformed_activations = [a.expand(b_n,*a.shape[1:]).clone() for a in untransformed_activations]
            # inverse transform selected activations
            self.activations_transformer.trasform_st_same_row(post_transformed_activations, t_start, t_end)
            t_start = t_end
            pre_transformed_activations = [a.cpu().numpy() for a in pre_transformed_activations]
            post_transformed_activations = [a.cpu().numpy() for a in post_transformed_activations]
            yield batch, pre_transformed_activations, post_transformed_activations

    def get_both_iterator(self) -> ActivationsIterator:
        return self

    def layer_names(self) -> [str]:
        return self.activations_transformer.layer_names




import torch
from torch.nn import functional as F
from typing import List, Iterator
from tmeasures.transformations.affine import RotationParameter,ScaleParameter,TranslationParameter
from tmeasures.transformations import affine
import numpy as np
import math

from tmeasures.pytorch.transformations import PyTorchTransformation, PyTorchTransformationSet


class AffineTransformation(affine.AffineTransformation,PyTorchTransformation):

    def __init__(self,ap:affine.AffineParameters):
        super().__init__(ap)
        self.transformation_matrix= self.get_transformation_matrix()
        self.transformation_matrix= self.transformation_matrix.unsqueeze(0)

    def get_transformation_matrix(self)->torch.Tensor:
        # center_matrix=torch.eye(3)
        # # center_matrix[:2,2]=-.5
        # decenter_matrix=torch.eye(3)
        # # decenter_matrix[:2,2]=.5

        r,s,t=self.ap
        r= r*180
        sx,sy=s
        #tx,ty=t
        matrix = torch.eye(3)
        matrix[:2, :2] = torch.tensor([[np.cos(r)/sx, -1.0*np.sin(r)/sy],
                                        [np.sin(r)/sx, np.cos(r)/sy]])
        matrix[:2,2]=torch.tensor(t)*2
        #matrix=decenter_matrix.mm(matrix.mm(center_matrix))
        # print(matrix)
        return matrix[:2,:]

    def __call__(self, x: torch.FloatTensor):
        with torch.no_grad(): # TODO is this necessary?

            c, h, w = x.shape
            x = torch.unsqueeze(x,0)
            shape=[1,c,h,w]
            grid = F.affine_grid(self.transformation_matrix,shape ,align_corners=False)
            grid=grid.to(x.device)
            x = F.grid_sample(x, grid,align_corners=False,padding_mode="border")
            return x[0,]

    def inverse(self):
        return AffineTransformation(self.ap.inverse())


class RotationTransformation(AffineTransformation):
    def __init__(self,p:affine.RotationParameter):
        self.p=p
        p=affine.AffineParameters(r=p)
        super().__init__(p)

    def parameters(self):
        # encode with cos and sin
        # https://arxiv.org/pdf/1702.01499.pdf0
        return torch.tensor([math.cos(self.p*360),math.sin(self.p*360)])

    def inverse(self):
        return RotationTransformation(-self.p)

class ScaleTransformation(AffineTransformation):
    def __init__(self,p:affine.ScaleParameter):
        self.p=p
        p=affine.AffineParameters(s=p)
        super().__init__(p)

    def parameters(self):
        sx,sy=self.p
        return torch.tensor([sx,sy])
    def inverse(self):
        sx,sy=self.p
        return ScaleTransformation((1/sx,1/sy))

class TranslationTransformation(AffineTransformation):
    def __init__(self,p:affine.TranslationParameter):
        self.p=p
        p=affine.AffineParameters(t=p)
        super().__init__(p)
    def parameters(self):
        tx, ty = self.p
        return torch.tensor([tx, ty])

    def inverse(self):
        tx,ty=self.p
        return TranslationTransformation((-tx,-ty))

class BaseAffineTransformationGenerator(affine.AffineGenerator,PyTorchTransformationSet):
    def __init__(self, r:Iterator[RotationParameter]=None, s:Iterator[ScaleParameter]=None, t:Iterator[TranslationParameter]=None):
        super().__init__(r,s,t)

class AffineGenerator(BaseAffineTransformationGenerator):
    def __init__(self, r:Iterator[RotationParameter]=None, s:Iterator[ScaleParameter]=None, t:Iterator[TranslationParameter]=None):
        super().__init__(r,s,t)

    def make_transformation(self, ap:affine.AffineParameters):
        return AffineTransformation(ap)

    def copy(self):
        return AffineGenerator(self.r,self.s,self.t)


class RotationGenerator(BaseAffineTransformationGenerator):
    def __init__(self, r:Iterator[RotationParameter]):
        super().__init__(r,None,None)

    def make_transformation(self, ap:affine.AffineParameters):
        return RotationTransformation(ap.r)

    def copy(self):
        return RotationGenerator(self.r)

class ScaleGenerator(BaseAffineTransformationGenerator):
    def __init__(self, s:Iterator[ScaleParameter]):
        super().__init__(None,s,None)

    def make_transformation(self, ap:affine.AffineParameters):
        return ScaleTransformation(ap.s)

    def copy(self):
        return ScaleGenerator(self.s)

class TranslationGenerator(BaseAffineTransformationGenerator):
    def __init__(self, t:Iterator[TranslationParameter]):
        super().__init__(None,None,t)

    def make_transformation(self, ap:affine.AffineParameters):
        return TranslationTransformation(ap.t)

    def copy(self):
        return TranslationGenerator(self.t)



"""
The :mod:`.np` module contains implementations of the measures for the NumPy framework, as well as some utility functions.

A :class:`.NumpyMeasure` can be evaluated with different adapters from PyTorch and Tensorflow, with corresponding models, datasets and transformations defined in those frameworks. The evaluation of the measure returns a :class:`.NumpyMeasureResult`

Variance-Based Invariance Measures
------------------------
* :class:`.np.SampleVarianceInvariance`
* :class:`.TransformationVarianceInvariance`
* :class:`.NormalizedVarianceInvariance`

Distance-Based Invariance Measures
------------------------
* :class:`.SampleDistanceInvariance`
* :class:`.TransformationDistanceInvariance`
* :class:`.NormalizedDistanceInvariance`

Other Invariance Measures
------------------------
* :class:`.GoodfellowInvariance`
* :class:`.ANOVAInvariance`
* :class:`.ANOVAFInvariance`

Variance-Based Same-Equivariance Measures
------------------------
* :class:`.SampleVarianceSameEquivariance`
* :class:`.TransformationVarianceSameEquivariance`
* :class:`.NormalizedVarianceSameEquivariance`

Distance-Based Same-Equivariance Measures
------------------------
* :class:`.SampleDistanceSameEquivariance`
* :class:`.TransformationDistanceSameEquivariance`
* :class:`.NormalizedDistanceSameEquivariance`

"""


from .base import NumpyMeasure
from .layer_transformation import MeasureTransformation, AggregateConvolutions, AggregateTransformation, \
    AggregateFunction, IdentityTransformation
from .quotient import QuotientMeasure, divide_activations

from .variance_invariance import NormalizedVarianceInvariance, SampleVarianceInvariance, \
    TransformationVarianceInvariance
from .anova import ANOVAInvariance, ANOVAFInvariance
from .distance_invariance import NormalizedDistanceInvariance, SampleDistanceInvariance, \
    TransformationDistanceInvariance, DistanceAggregation

from .distance_same_equivariance import NormalizedDistanceSameEquivariance, TransformationDistanceSameEquivariance, \
    SampleDistanceSameEquivariance

from .variance_same_equivariance import NormalizedVarianceSameEquivariance, TransformationVarianceSameEquivariance, \
    SampleVarianceSameEquivariance

from .same_equivariance_simple import DistanceSameEquivarianceSimple, DistanceFunction

from .goodfellow import GoodfellowInvariance
from .goodfellow_prob import GoodfellowNormalInvariance, GoodfellowNormalLocalInvariance, \
    GoodfellowNormalGlobalInvariance




from tmeasures.transformation import Transformation,TransformationSet
from typing import Tuple,Iterator
import itertools
import abc
import torch
TranslationParameter = Tuple[float, float]
ScaleParameter = Tuple[float, float]
RotationParameter = float


class AffineParameters:
    def __init__(self,r:RotationParameter=0,s:ScaleParameter=(1,1),t:TranslationParameter=(0,0)):
        self.r=r
        self.s=s
        self.t=t
    def __iter__(self):
        return iter((self.r,self.s,self.t))

    def inverse(self):
        r, s, t = self.r,self.s,self.t
        r=-r
        tx,ty=t
        t= (-tx,-ty)
        sx,sy=s
        s=(1/sx,1/sy)
        return AffineParameters(r,s,t)

    def as_tensor(self):
        r,(sx,sy),(tx,ty) = self.r,self.s,self.t
        return torch.tensor([r,sx,sy,tx,ty])

    def __repr__(self):
        r, s, t = self
        return f"AffineParameters(r={r},t={t},s={s})"

class AffineTransformation(Transformation):
    def __init__(self,ap:AffineParameters):
        self.ap=ap

    def __eq__(self, other):
        if self.__class__ == other.__class:
            return self.ap == other.parameters
        else:
            return False

    def __repr__(self):
        r, s, t = self.ap
        return f"AffineTransformation(r={r},s={s},t={t})"

    @abc.abstractmethod
    def inverse(self):
        pass
    def parameters(self):
        return self.ap.as_tensor()

def ifnone(x,v):
    if x is None:
        return v
    else:
        return x

from .parameters import NoRotation,NoScale,NoTranslation

class AffineGenerator(TransformationSet):
    def __init__(self, r:[RotationParameter]=None, s:[ScaleParameter]=None, t:[TranslationParameter]=None):
        r=ifnone(r,NoRotation())
        s=ifnone(s,NoScale())
        t=ifnone(t,NoTranslation())
        self.r,self.s,self.t=r,s,t
        parameters = itertools.product(r,s,t)
        parameters = [AffineParameters(r,s,t) for r,s,t in parameters]
        self.transformations = [self.make_transformation(p) for p in parameters]
        super().__init__(self.transformations)

    def valid_input(self,shape:Tuple[int, ]) -> bool:
        return len(shape) == 4

    @abc.abstractmethod
    def make_transformation(self, ap:AffineParameters)->AffineTransformation:
        pass

    def __repr__(self):
        return f"Affine(r={self.r},s={self.s},t={self.t})"

    # def __eq__(self, other):
    #     if isinstance(other,self.__class__):
    #         other:SimpleAffineTransformationGenerator = other
    #         return self.r == other.r and self.s==other.s and self.t==other.t

    def id(self):
        return str(self)

    def inverse(self)->[Transformation]:
        return [t.inverse() for t in self]




from ..base import NumpyMeasure
import numpy as np
from ..layer_transformation import MeasureTransformation
from ..stats_running import RunningMeanAndVarianceWelford,RunningMeanWelford
import tmeasures as tm
from multiprocessing import Queue
from ..quotient import QuotientMeasure
from .functions import MeasureFunction
from .multithreaded_layer_measure import LayerMeasure,PerLayerMeasure,ActivationsOrder

class VarianceLayerMeasure(LayerMeasure):

    def __init__(self, id:int, name:str, measure_function: MeasureFunction, conv_aggregation:MeasureTransformation):
        super().__init__(id,name)
        self.conv_aggregation=conv_aggregation
        self.measure_function=measure_function

    def eval(self,q:Queue,inner_q:Queue):

        m = RunningMeanWelford()
        for iteration_info in self.queue_as_generator(q):
            inner_m = RunningMeanAndVarianceWelford()

            for activations in self.queue_as_generator(inner_q):
                activations = self.conv_aggregation.apply(activations)
                inner_m.update_all(activations)
                # i += 1

            inner_result = self.measure_function.apply_running(inner_m)
            m.update(inner_result)

        return m.mean()

class VarianceMeasure(PerLayerMeasure):
    def __init__(self, order:ActivationsOrder, measure_function: MeasureFunction, conv_aggregation: MeasureTransformation):
        super().__init__(order)
        self.measure_function = measure_function
        self.conv_aggregation = conv_aggregation

    def generate_layer_measure(self, i:int, name:str) -> LayerMeasure:
        return VarianceLayerMeasure(i,name,self.measure_function,self.conv_aggregation)

class TransformationVarianceMeasure(VarianceMeasure):
    def __init__(self, measure_function: MeasureFunction, conv_aggregation: MeasureTransformation):
        super().__init__(ActivationsOrder.SamplesFirst,measure_function,conv_aggregation)
    def __repr__(self):
        return f"TVM(f={self.measure_function.value},ca={self.conv_aggregation.value})"

class SampleVarianceMeasure(VarianceMeasure):
    def __init__(self, measure_function: MeasureFunction, conv_aggregation: MeasureTransformation):
        super().__init__(ActivationsOrder.TransformationsFirst,measure_function,conv_aggregation)

    def __repr__(self):
        return f"SVM(f={self.measure_function.value},ca={self.conv_aggregation.value})"

class NormalizedVarianceMeasure(QuotientMeasure):

    def __init__(self, measure_function: MeasureFunction, conv_aggregation:MeasureTransformation):
        sm = SampleVarianceMeasure(measure_function, conv_aggregation)
        ttm = TransformationVarianceMeasure(measure_function, conv_aggregation)
        super().__init__(ttm,sm)
        self.numerator_measure = ttm
        self.denominator_measure = sm
        self.measure_function = measure_function
        self.conv_aggregation = conv_aggregation

    def __repr__(self):
        return f"NVM(f={self.measure_function.value},ca={self.conv_aggregation.value})"


from __future__ import annotations

def indices_of(list:[],value)->[int]:
    indices =[]
    for i,l in enumerate(list):
        if value == l:
            indices.append(i)
    return indices

def get_all(list:[],indices:[int])->[]:
    return [list[i] for i in indices]




from queue import Queue
from typing import Sized,Iterable

class IterableQueue(Queue,Sized,Iterable):
    """Queue supporting Iterator and Sized protocols.
    Queue has a max size so it can be iterated upon with a for loop
    """
    def __init__(self, n:int,maxsize=None):
        if maxsize is None:
            maxsize=n
        super().__init__(maxsize)
        self.n=n
        self.i=0

    def __len__(self):
        return self.n


    def __iter__(self):
        return self

    def __next__(self):
        self.i+=1
        if self.i == self.n:
            raise StopIteration()
        return self.get()


import cv2
import numpy as np
from skimage import transform as skimage_transform

from ..affine import AffineTransformation,AffineParameters


class AffineTransformationNumpy(AffineTransformation):
    def __init__(self,p:AffineParameters):
        super().__init__(p)
        self.transformation_matrix=self.generate_transformation_matrix(p)

    def generate_transformation_matrix(self, p:AffineParameters):
        transformation = skimage_transform.AffineTransform(scale=p.s, rotation=p.r*180, shear=None,translation=p.t)
        return transformation.params

    def center_transformation(self,transformation:skimage_transform.AffineTransform,image_size):
        h,w=image_size
        shift_y, shift_x = (h- 1) / 2., (w- 1) / 2.
        shift = skimage_transform.AffineTransform(translation=[-shift_x, -shift_y])
        shift_inv = skimage_transform.AffineTransform(translation=[shift_x, shift_y])
        return shift + (transformation+ shift_inv)

    def single(self,image:np.ndarray)->np.ndarray:
        input_shape=image.shape
        transformation= self.center_transformation(self.transformation_matrix,input_shape[:2])
        image_size=tuple(input_shape[:2])
        if input_shape[2] == 1:
            image = image[:, :, 0]
        image= cv2.warpPerspective(image, transformation, image_size)
        if input_shape[2]==1:
           image= image[:, :, np.newaxis]
        return image

    def __call__(self, batch:np.ndarray)->np.ndarray:

        results=[]
        for i in range(batch.shape[0]):
            x= batch[i, :]
            x=self.single(x)
            results.append(x)
        return np.stack(results, axis=0)

    def inverse(self):
        return AffineTransformationNumpy(self.ap.inverse())

    def numpy(self):
        return self


import numpy as np
from abc import ABC,abstractmethod
from enum import Enum
from .. import MeasureResult
from typing import Tuple

class MeasureTransformation(ABC):

    @abstractmethod
    def apply(self,r:MeasureResult):
        pass
    def __repr__(self):
        return self.__class__.__name__

class IdentityTransformation(MeasureTransformation):
    def apply(self,r:MeasureResult):
        return r


class AggregateFunction(Enum):
    mean = "mean"
    max = "max"
    sum = "sum"

    def functions(self):
        return {AggregateFunction.mean: np.nanmean
            , AggregateFunction.sum: np.nansum
            , AggregateFunction.max: np.nanmax
                }
    def get_function(self):
        return self.functions()[self]
    def __repr__(self):
        return self.value


class AggregateTransformation(MeasureTransformation):

    def __init__(self, f:AggregateFunction=AggregateFunction.mean, axis:Tuple[int]=None):
         self.f = f
         if axis is None:
            self.axis=(0,) # keep only axis 0
         else:
            assert len(axis) > 0
            self.axis = axis # axis to keep

    def apply(self, r:MeasureResult)->MeasureResult:
        aggregate_function= self.f.get_function()
        results=[]
        for layer in r.layers:
            # only apply to layers with at least max(self.axis) dims
            dims= len(layer.shape)
            if max(self.axis) < dims:
                diff = set(range(dims)).difference(set(self.axis))
                diff = tuple(diff)
                flat_activations = aggregate_function(layer,axis=diff)
                # assert len(flat_activations.shape) == len(self.axis),f"After collapsing, the activation shape should have only {len(self.axis)} dimensions. Found vector with shape {flat_activations.shape} instead."
            else:
                flat_activations = layer.copy()
            results.append(flat_activations)

        return MeasureResult(results,r.layer_names,r.measure,r.extra_values)
    def __repr__(self):
        return f"{self.__class__.__name__}(f={self.f.value},axis={self.axis})"




# def apply(self, layer:np.ndarray) -> np.ndarray:
#     '''
#     :param layer:  a 4D np array (else apply no aggregation)
#     :return:
#     '''
#
#     # none does no aggregation
#     if self == ConvAggregation.none:
#         return layer
#     layer=np.abs(layer)
#     # only aggregate conv layers (n,c,h,w)
#     if len(layer.shape) != 4:
#         return layer
#
#     function = self.functions()[self]
#     n, c, h, w = layer.shape
#     flat_activations = np.zeros((n, c))
#     for i in range(n):
#         flat_activations[i, :] = function(layer[i, :, :, :], axis=(1,2))
#
#     return flat_activations
#

class AggregateConvolutions(AggregateTransformation):
    def __init__(self, f:AggregateFunction=AggregateFunction.mean):
        super().__init__(f,(0,))



from typing import Dict, List
from tmeasures import MeasureResult
from .activations_iterator import ActivationsIterator
from . import NumpyMeasure

from multiprocessing import Queue
from .quotient import divide_activations
from .multithread.multithreaded_layer_measure import LayerMeasure,PerLayerMeasure,ActivationsOrder
import numpy as np
from .stats_running import RunningMeanWelford

class GoodfellowGlobalInvariance(LayerMeasure):

    def __init__(self, id:int, name:str, activation_percentage:float,sign:int):
        super().__init__(id,name)
        self.activation_percentage = activation_percentage
        self.sign=sign

    def eval(self,q:Queue,inner_q:Queue):

        max_samples_per_transformation=20
        history=[]
        print(f"starting {self.id}")
        for transformation in self.queue_as_generator(q):
            i=0
            for activations in self.queue_as_generator(inner_q):
                if i<max_samples_per_transformation:
                    history.append(activations)
                i+=activations.shape[0]

        print(f"finish{self.id}")
        # sort all activation values for the different samples
        activations = np.vstack(history)
        del history
        if self.sign != 1:
            activations*=self.sign
        print(f"shape {activations.shape}")
        activations.sort(axis=0)
        n = activations.shape[0]
        # calculate the threshold indexes so that G(i) = activation_percentage
        threshold_indexes = int((1-self.activation_percentage)*n)
        # calculate the threshold values (approximately)
        self.thresholds = activations[threshold_indexes, :]

        # set g(i) equal to the activations_percentage
        self.g= np.zeros_like(self.thresholds) + self.activation_percentage


    def get_final_result(self):
        return self.g

class GoodfellowLocalInvariance(LayerMeasure):

    def __init__(self, id:int, name:str, threshold:float,sign:int):
        super().__init__(id,name)
        self.threshold = threshold
        self.sign=sign

    def eval(self,q:Queue,inner_q:Queue):
        running_mean = RunningMeanWelford()
        # activation_sum=0
        n=0
        for transformation in self.queue_as_generator(q):
            for activations in self.queue_as_generator(inner_q):
                if self.sign != 1:
                    activations *= self.sign
                activated = (activations > self.threshold) * 1
                running_mean.update_all(activated)

        self.l=running_mean.mean()

    def get_final_result(self):
        return self.l


class GlobalFiringRateMeasure(PerLayerMeasure):
    def __init__(self,activation_percentage:float,sign:int):
        super().__init__(ActivationsOrder.TransformationsFirst)
        self.activation_percentage:float=activation_percentage
        self.sign:int=sign
        self.layer_measures:Dict[int, GoodfellowGlobalInvariance]={}

    def __repr__(self):
        return f"G()"

    def generate_layer_measure(self, i:int, name:str) -> LayerMeasure:
        layer_measure = GoodfellowGlobalInvariance(i, name, self.activation_percentage, self.sign)
        self.layer_measures[i]=layer_measure
        return layer_measure

    def get_layer_measure(self,i:int):
        return self.layer_measures[i]

    def get_thresholds(self):
        return {i:m.thresholds  for i,m in self.layer_measures.items()}

class LocalFiringRateMeasure(PerLayerMeasure):
    def __init__(self,thresholds:List[np.ndarray],sign:int):
        super().__init__(ActivationsOrder.SamplesFirst)
        self.sign=sign
        self.thresholds=thresholds

    def __repr__(self):
        return f"L()"

    def generate_layer_measure(self, i:int, name:str) -> LayerMeasure:

        return GoodfellowLocalInvariance(i, name, self.thresholds[i], self.sign)

class GoodfellowInvariance(NumpyMeasure):


    def __init__(self,activations_percentage=0.01,sign=1):
        assert sign in [1,-1]
        super().__init__()
        self.activations_percentage=activations_percentage
        self.sign=sign

    def eval(self,activations_iterator:ActivationsIterator):
        print("calculating global")
        self.g = GlobalFiringRateMeasure(self.activations_percentage,self.sign)
        g_result = self.g.eval(activations_iterator)
        print("calculating thresholds")
        self.thresholds = self.g.get_thresholds()
        print("calculating local")
        self.l = LocalFiringRateMeasure(self.thresholds,self.sign)
        l_result = self.l.eval(activations_iterator)

        ratio = divide_activations(l_result.layers,g_result.layers)
        return MeasureResult(ratio, activations_iterator.layer_names(), self)

    def __repr__(self):
        return f"Goodfellow(gp={self.activations_percentage})"

    def name(self):
        return "Goodfellow"
    def abbreviation(self):
        return "GF"


from tmeasures import MeasureResult
from ..base import NumpyMeasure
from ..activations_iterator import ActivationsIterator
import numpy as np
from ..stats_running import RunningMeanWelford

from abc import abstractmethod

from typing import Callable
class EventDrivenLayerMeasure():

    def __init__(self,layer_index:int,layer_name:str):
        self.layer_name=layer_name
        self.layer_index=layer_index
    @abstractmethod
    def update_layer(self, activations:np.ndarray):
        pass

    @abstractmethod
    def get_final_result(self)->np.ndarray:
        pass
    def on_begin_iteration(self):
        pass
    def on_begin(self):
        pass

class LayerTransformationMeasure(NumpyMeasure):
    def __init__(self, layer_measure_generator:Callable[[int,str], EventDrivenLayerMeasure]):
        super().__init__()
        self.layer_measure_generator=layer_measure_generator

    def eval(self,activations_iterator:ActivationsIterator)->MeasureResult:
        layer_names=activations_iterator.layer_names()
        n_intermediates = len(layer_names)
        layer_measures = [self.layer_measure_generator(i,n) for i,n in enumerate(layer_names)]

        for r in layer_measures:
            r.on_begin()

        for activations, x_transformed in activations_iterator.samples_first():
            for r in layer_measures:
                r.on_begin_iteration()
            # activations has the activations for all the transformations
            for j, layer_activations in enumerate(activations):
                layer_measures[j].update_layer(layer_activations)

        results = [r.get_final_result() for r in layer_measures]
        return MeasureResult(results,layer_names,self)

#
# class LayerSampleMeasure(Measure):
#     def __init__(self, layer_measure_generator:Callable[[int,str], EventDrivenLayerMeasure]):
#         super().__init__()
#         self.layer_measure_generator=layer_measure_generator
#
#     def __repr__(self):
#         return f"(f={self.measure_function.value},da={self.distance_aggregation.value})"
#
#     def eval(self,activations_iterator:ActivationsIterator)->MeasureResult:
#         layer_names = activations_iterator.activation_names()
#         n_layers = len(layer_names)
#         mean_variances_running = [RunningMean() for i in range(n_layers)]
#
#         for transformation, transformation_activations in activations_iterator.transformations_first():
#             # calculate the variance of all samples for this transformation
#             for x, batch_activations in transformation_activations:
#                 for j, layer_activations in enumerate(batch_activations):
#                     for i in range(layer_activations.shape[0]):
#                         layer_measure = self.measure_function.apply(layer_activations)
#                         mean_variances_running [j].update(layer_measure )
#
#         # calculate the final mean over all transformations (and layers)
#         mean_variances = [b.mean() for b in mean_variances_running]
#         return MeasureResult(mean_variances,layer_names,self)


class STDLayerMeasure(EventDrivenLayerMeasure):

    def __init__(self,layer_index:int,layer_name:str):
        super(STDLayerMeasure).__init__(layer_index,layer_name)
        self.running_mean = RunningMeanWelford()

    def update_layer(self, activations):
        layer_measure = activations.std(axis=0)
        self.running_mean.update(layer_measure)

    def get_final_result(self):
        return self.running_mean.mean()



import tmeasures as tm
from torch.utils.data import Dataset
from .base import PyTorchMeasure, PyTorchMeasureOptions, PyTorchMeasureResult
from .activations_iterator import PytorchActivationsIterator
from . import ActivationsModule
from .layer_measures import Variance
from .quotient import QuotientMeasure
from .measure_transformer import MeasureTransformation, NoTransformation


class TransformationVarianceInvariance(PyTorchMeasure):

    def eval(self, dataset: Dataset, transformations: tm.TransformationSet, model: ActivationsModule,
             o: PyTorchMeasureOptions):
        dataset2d = tm.pytorch.dataset2d.SampleTransformationDataset(dataset, transformations, device=o.data_device)
        iterator = PytorchActivationsIterator(model, dataset2d, o)
        results = iterator.evaluate(Variance())
        return PyTorchMeasureResult(results, model.activation_names(), self)
    

class SampleVarianceInvariance(PyTorchMeasure):

    def eval(self, dataset: Dataset, transformations: tm.TransformationSet, model: ActivationsModule,
             o: PyTorchMeasureOptions):
        dataset2d = tm.pytorch.dataset2d.TransformationSampleDataset(dataset, transformations, device=o.data_device)
        iterator = PytorchActivationsIterator(model, dataset2d, o)
        results = iterator.evaluate(Variance())
        return PyTorchMeasureResult(results, model.activation_names(), self)


class NormalizedVarianceInvariance(QuotientMeasure):

    def __init__(self, measure_transformation: MeasureTransformation = NoTransformation()):
        super().__init__(TransformationVarianceInvariance(), SampleVarianceInvariance(),
                         measure_transformation=measure_transformation)

    def __repr__(self):
        return f"{self.abbreviation()}({self.measure_transformation})"



from . import PyTorchMeasure, ActivationsByLayer, PyTorchMeasureOptions, ActivationsModule, PyTorchMeasureResult
from .. import MeasureResult, Measure
from torch.utils.data import Dataset
import torch
from .. import TransformationSet
from .measure_transformer import MeasureTransformation, NoTransformation
import tqdm.auto as tqdm


def safe_divide(x: torch.Tensor, y: torch.Tensor):
    eps = 0
    r = x.clone()

    r[y > eps] /= y[y > eps]
    both_below_eps = torch.logical_and(y <= eps,
                                       x <= eps)
    r[both_below_eps] = 1

    only_baseline_below_eps = torch.logical_and(y <= eps, x > eps)
    # print("num", np.where(num_values > eps))
    # print("den", np.where(den_values <= eps))
    # print("both", np.where(only_baseline_below_eps))

    r[only_baseline_below_eps] = float("Inf")
    return r


def divide_activations(num: ActivationsByLayer, den: ActivationsByLayer) -> ActivationsByLayer:
    # TODO evaluate other implementations
    # https://stackoverflow.com/questions/26248654/how-to-return-0-with-divide-by-zero
    return [safe_divide(x, y) for (x, y) in zip(num, den)]


class QuotientMeasureResult(PyTorchMeasureResult):
    def __init__(self, v: ActivationsByLayer, layer_names: [str], m: Measure, x_result: PyTorchMeasureResult,
                 y_result: PyTorchMeasureResult):
        super().__init__(v, layer_names, m)
        self.x_result = x_result
        self.y_result = y_result

    def numpy(self):
        self_numpy=super().numpy()
        self_numpy.x_result = self.x_result.numpy()
        self_numpy.y_result = self.y_result.numpy()
        return self_numpy


class QuotientMeasure(PyTorchMeasure):
    def __init__(self, numerator_measure: PyTorchMeasure, denominator_measure: PyTorchMeasure,
                 measure_transformation: MeasureTransformation = NoTransformation()):
        super().__init__()
        self.numerator_measure = numerator_measure
        self.denominator_measure = denominator_measure
        self.measure_transformation = measure_transformation

    def __repr__(self):
        return f"QM({self.numerator_measure}_DIV_{self.denominator_measure},{self.measure_transformation})"

    def eval(self, dataset: Dataset, transformations: TransformationSet, model: ActivationsModule,
             o: PyTorchMeasureOptions) -> QuotientMeasureResult:
        with tqdm.tqdm(total=2, disable=not o.verbose, colour='green') as pbar:
            v_transformations = self.numerator_measure.eval(dataset, transformations, model, o)
            self.measure_transformation.transform_result(v_transformations)
            pbar.update(1)
            v_samples = self.denominator_measure.eval(dataset, transformations, model, o)
            self.measure_transformation.transform_result(v_samples)
            pbar.update(2)
        v = [safe_divide(x, y) for (x, y) in zip(v_transformations.layers, v_samples.layers)]

        return QuotientMeasureResult(v, model.activation_names(), self, v_transformations, v_samples)



import tmeasures as tm
from torch.utils.data import Dataset
from .base import PyTorchMeasure, PyTorchMeasureOptions, PyTorchMeasureResult
from .activations_iterator import PytorchActivationsIterator
from . import ActivationsModule, Variance
from .. import InvertibleTransformation
from .quotient import QuotientMeasure
import torch

from .activations_iterator import ActivationsTransformer
from .measure_transformer import MeasureTransformation, NoTransformation


class InverseTransformationTransformer(ActivationsTransformer):

    def transform(self, activations: torch.Tensor, x: torch.Tensor,
                  transformations: [InvertibleTransformation]) -> torch.Tensor:
        transformed = []
        for i in range(activations.shape[0]):
            inverse = transformations[i].inverse()
            transformed.append(inverse(activations[i,]))
        return torch.stack(transformed, dim=0)


class TransformationVarianceSameEquivariance(PyTorchMeasure):

    def eval(self, dataset: Dataset, transformations: tm.TransformationSet, model: ActivationsModule,
             o: PyTorchMeasureOptions):
        dataset2d = tm.pytorch.dataset2d.SampleTransformationDataset(dataset, transformations, device=o.data_device)
        iterator = PytorchActivationsIterator(model, dataset2d, o,
                                              activations_transformer=InverseTransformationTransformer())
        results = iterator.evaluate(Variance())
        return PyTorchMeasureResult(results, model.activation_names(), self)


class SampleVarianceSameEquivariance(PyTorchMeasure):

    def eval(self, dataset: Dataset, transformations: tm.TransformationSet, model: ActivationsModule,
             o: PyTorchMeasureOptions):
        dataset2d = tm.pytorch.dataset2d.TransformationSampleDataset(dataset, transformations, device=o.data_device)
        iterator = PytorchActivationsIterator(model, dataset2d, o,
                                              activations_transformer=InverseTransformationTransformer())
        results = iterator.evaluate(Variance())
        return PyTorchMeasureResult(results, model.activation_names(), self)


class NormalizedVarianceSameEquivariance(QuotientMeasure):

    def __init__(self, measure_transformation: MeasureTransformation = NoTransformation()):
        super().__init__(TransformationVarianceSameEquivariance(), SampleVarianceSameEquivariance(),
                         measure_transformation=measure_transformation)

    def __repr__(self):
        extra =""
        return f"{self.abbreviation()}({self.measure_transformation})"



import torch
import abc


class RunningMeasure(abc.ABC):
    @abc.abstractmethod
    def clear(self):
        pass

    @abc.abstractmethod
    def update(self, x: torch.Tensor):
        pass

    def update_all(self, x: torch.Tensor):
        for i in range(x.shape[0]):
            self.update(x[i, :])

    @abc.abstractmethod
    def update_batch(self, x: torch.Tensor):
        pass


class RunningMean(RunningMeasure):

    @abc.abstractmethod
    def mean(self):
        pass


class RunningVariance(RunningMeasure):
    @abc.abstractmethod
    def var(self):
        pass

    def std(self):
        return torch.sqrt(self.var())


class RunningMeanWelford(RunningMean):
    def __repr__(self):
        return f"RunningMean(n={self.n},mean={self.mean().shape})"

    def __init__(self):
        self.clear()

    def clear(self):
        self.n = 0
        self.mu = 0

    def update(self, x: torch.Tensor):
        self.n += 1
        self.mu += (x - self.mu) / self.n

    def update_all(self, x: torch.Tensor):
        self.update_batch(x)

    def update_batch(self, x: torch.Tensor):
        k = x.shape[0]
        self.n += k
        diff = x - self.mu
        self.mu += diff.sum(dim=0) / self.n

    def mean(self):
        return self.mu if self.n > 0 else torch.zeros(1)


class RunningMeanAndVarianceWelford(RunningMean, RunningVariance):

    def __repr__(self):
        return f"RunningMeanAndVarianceWellford(n={self.n},mean={self.mean().shape})"

    def __init__(self):
        self.clear()

    def clear(self):
        self.n = 0
        self.m = 0
        self.s = 0

    def update(self, x):
        # x = x.double()
        self.n += 1
        diff = x - self.m
        self.m = self.m + (diff / self.n)
        self.s = self.s + diff * (x - self.m)

    def update_batch(self, x: torch.Tensor):
        # x = x.double()
        # see https://hackmd.io/_-rPOUx5RpajaZZkyU9wgw
        k = x.shape[0]
        self.n += k
        diff = x - self.m
        self.m += diff.sum(dim=0) / self.n
        diff2 = (diff * (x - self.m)).sum(dim=0)
        self.s += diff2

    def mean(self):
        # assert(self.n>0)
        return self.m if self.n else torch.zeros_like(self.m)

    def var(self):
        # assert (self.n > 1)
        return self.s / (self.n - 1) if self.n > 1 else torch.zeros_like(self.s)



class RunningMeanSets(RunningMean):
    def __repr__(self):
        return f"RunningMeanSets(n={self.n},mean={self.mean().shape})"

    def __init__(self):
        self.clear()

    def clear(self):
        self.n = 0
        self.mu = 0

    def mean(self):
        return self.mu

    def update_batch(self, x: torch.Tensor):
        m = x.shape[0]
        mu_x = x.mean(dim=0)
        n = self.n
        self.n = m + n
        if n == 0:
            self.mu = mu_x
        else:
            c1, c2 = n / self.n, m / self.n
            self.mu = c1 * self.mu + c2 * mu_x

    def update(self, x: torch.Tensor):
        self.n += 1
        if self.n == 1:
            self.mu = x
        else:
            self.mu = self.mu + (x - self.mu) / self.n


# noinspection PyAttributeOutsideInit
class RunningMeanVarianceSets(RunningMean, RunningVariance):
    def __repr__(self):
        return f"RunningMeanVarianceSets(n={self.n},mean={self.mean().shape})"

    def __init__(self):
        self.clear()

    def clear(self):
        self.n = 0
        self.mu = 0
        self.v = 0

    def update_batch(self, x: torch.Tensor):

        m = x.shape[0]
        mu_x = x.mean(dim=0)
        if m>1:
            v_x = x.var(dim=0, unbiased = False)
        else:
            v_x=0
        n = self.n
        self.n = m + n
        if n == 0:
            self.mu = mu_x
            self.v = v_x
        else:
            c1, c2 = n / self.n, m / self.n
            c3 = c1*c2
            # print(c1,c2,c3)
            mu = self.mu
            self.mu = c1 * self.mu + c2 * mu_x
            self.v = c1 * self.v + c2 * v_x + c3 * (mu - mu_x) ** 2
            # print(self.v)

    def var(self):
        return self.v*self.n/(self.n-1)

    def mean(self):
        return self.mu

    def update(self, x: torch.Tensor):
        self.update_batch(torch.unsqueeze(x,dim=0))




from __future__ import annotations

from typing import List
from .activations_iterator import ActivationsIterator
from tmeasures import MeasureResult
from . import NumpyMeasure,MeasureTransformation
from .stats_running import RunningMeanWelford
from .quotient import divide_activations

from .distance_invariance import DistanceAggregation

def list_get_all(list:List,indices:List[int])->List:
    return [list[i] for i in indices]

class BaseDistanceSameEquivarianceMeasure(NumpyMeasure):
    def __init__(self, distance_aggregation:DistanceAggregation):
        super().__init__()
        self.distance_aggregation=distance_aggregation

class TransformationDistanceSameEquivariance(BaseDistanceSameEquivarianceMeasure):
    def __repr__(self):
        return f"TDSE(da={self.distance_aggregation})"

    def name(self)->str:
        return "Transformation Distance Same-Equivariance"
    def abbreviation(self):
        return "TDSE"


    def eval(self,activations_iterator:ActivationsIterator,verbose=False)->MeasureResult:
        activations_iterator = activations_iterator.get_inverted_activations_iterator()
        mean_running= None

        for x, transformation_activations_iterator in activations_iterator.samples_first():
            # transformation_activations_iterator can iterate over all transforms
            for x_transformed, activations in transformation_activations_iterator:
                if mean_running is None:
                    mean_running = [RunningMeanWelford() for i in range(len(activations))]
                for j, layer_activations in enumerate(activations):
                    layer_measure= self.distance_aggregation.apply(layer_activations)
                    # update the mean over all transformation
                    mean_running[j].update(layer_measure)
        # calculate the final mean over all samples (and layers)
        means = [b.mean() for b in mean_running]
        return MeasureResult(means,activations_iterator.layer_names(),self)



class SampleDistanceSameEquivariance(BaseDistanceSameEquivarianceMeasure):

    def __repr__(self):
        return f"SDSE(da={self.distance_aggregation})"

    def name(self)->str:
        return "Sample Distance Same-Equivariance"
    def abbreviation(self):
        return "SDSE"


    def eval(self,activations_iterator:ActivationsIterator,verbose=False)->MeasureResult:
        activations_iterator = activations_iterator.get_inverted_activations_iterator()
        mean_running= None

        for transformation, samples_activations_iterator in activations_iterator.transformations_first():

            # transformation_activations_iterator can iterate over all transforms
            for x,activations in samples_activations_iterator:
                if mean_running is None:
                    mean_running = [RunningMeanWelford() for i in range(len(activations))]
                for j, layer_activations in enumerate(activations):
                    layer_measure= self.distance_aggregation.apply(layer_activations)
                    # update the mean over all transformation
                    mean_running[j].update(layer_measure)
        # calculate the final mean over all samples (and layers)
        means = [b.mean() for b in mean_running]
        return MeasureResult(means,activations_iterator.layer_names(),self)


class NormalizedDistanceSameEquivariance(NumpyMeasure):
    def __init__(self, distance_aggregation:DistanceAggregation):
        super().__init__()
        self.distance_aggregation=distance_aggregation
        self.transformation_measure=TransformationDistanceSameEquivariance(distance_aggregation)
        self.sample_measure=SampleDistanceSameEquivariance(distance_aggregation)


    def __repr__(self):
        return f"NDSE(da={self.distance_aggregation})"

    def name(self)->str:
        return "Normalized Distance Same-Equivariance"
    def abbreviation(self):
        return "NDSE"
    transformation_key=TransformationDistanceSameEquivariance.__name__
    sample_key=SampleDistanceSameEquivariance.__name__

    def eval(self,activations_iterator:ActivationsIterator,verbose=False) ->MeasureResult:

        transformation_result = self.transformation_measure.eval(activations_iterator,verbose)
        sample_result = self.sample_measure.eval(activations_iterator,verbose)
        result=divide_activations(transformation_result.layers,sample_result.layers)

        extra_values={ self.transformation_key:transformation_result,
                       self.sample_key:sample_result,
                       }
        return MeasureResult(result, transformation_result.layer_names,self,extra_values=extra_values)



from __future__ import annotations

import enum
import tmeasures as tm
import matplotlib.pyplot as plt
import numpy as np
from typing import List
from pathlib import Path
from tmeasures.np.stats_running import  RunningMeanAndVarianceWelford
import tmeasures as tm
from matplotlib.lines import Line2D

from scipy.stats import f_oneway, anderson_ksamp

from statsmodels.stats.multicomp import pairwise_tukeyhsd

def tukey_same_mean_test(results:List[tm.MeasureResult],alpha:float,max_samples_per_group=2000):
    n_layers = [len(r.layers) for r in results]
    n = n_layers[0]
    assert all([x==n for x in n_layers]),"Results should have the same number of activations"
    rejections = np.zeros(n)
    for i in range(n):
        values = [r.layers[i].flatten() for r in results]
        for j,v in enumerate(values):
            if v.size >max_samples_per_group:
                values[j]=v[np.random.choice(v.size,max_samples_per_group,replace=False)]
        groups = np.concatenate([np.zeros(len(v))+i for i,v in enumerate(values)])
        values = np.concatenate((values))
        tukey = pairwise_tukeyhsd(values,groups,alpha=alpha)
        rejections[i]=tukey.reject.mean()

    return rejections

def anderson_ksamples_test(measures:List[tm.MeasureResult]):
    n_layers = [len(r.layers) for r in measures]
    n = n_layers[0]
    assert all(map(lambda x: x==n,n_layers)),"Results should have the same number of activations"
    results = np.zeros(n)
    for i in range(n):
        groups = [r.layers[i].flatten() for r in measures]
        f,critical,significance = anderson_ksamp()(*groups)
        results[i]=significance
        

    return results

def anova_same_mean_test(measures:List[tm.MeasureResult]):
    n_layers = [len(r.layers) for r in measures]
    n = n_layers[0]
    assert all(map(lambda x: x==n,n_layers)),"Results should have the same number of activations"
    p_values = np.zeros(n)
    for i in range(n):
        groups = [r.layers[i].flatten() for r in measures]
        f,p_value = f_oneway(*groups)
        p_values[i]=p_value
        

    return p_values


default_y_lim=1.4

def plot_average_activations_different_models(results:List[tm.MeasureResult], labels:List[str]=None, title:str="", linestyles=None, colors=None, legend_location=None, markers:List[List[int]]=None, ylim=None):
    if ylim is None:
        ylim = default_y_lim
    f=plt.figure(dpi=300)
    n = len(results)
    if n == 0:
        raise ValueError(f"`results` is an empty list.")
    colors = get_colors(colors, n)

    f, ax = plt.subplots(dpi=get_dpi(n))
    f.suptitle(title)

    if linestyles is None and n<=4:
        linestyles=["-","--",":","-."]

    result_layers = np.array([len(r.layer_names) for r in results])
    min_n, max_n = result_layers.min(), result_layers.max()
    max_value=0
    
    for i, result in enumerate(results):
        n_layers = len(result.layers)
        x = np.linspace(0,100,n_layers,endpoint=True)
       
        y = result.per_layer_average()
        max_value = max(max_value, y.max())
        if labels is None:
            label = None
        else:
            label = labels[i]
        linestyle = get_default(linestyles,i,"-")

        color=colors[i, :]
        ax.plot(x, y, label=label, linestyle=linestyle, color=color,marker="o",markersize=3)

        if not markers is None:
            mark_layers = markers[i]
            x_mark = x[mark_layers]
            y_mark = y[mark_layers]
            ax.plot(x_mark, y_mark, linestyle="", color=color, marker="s")

    ax.set_ylabel("Measure")
    ax.set_xlabel(f"Activation (%)")
    ax.set_ylim(0, max(max_value * 1.1, ylim))

    ticks = list(range(0, 110, 10,))
    ax.set_xticks(ticks)


    box = ax.get_position()
    ax.set_position([box.x0, box.y0 + box.height * 0.1,
                     box.width, box.height * 0.9])


    if not labels is None:
        # Put legend below current axis
        handles, labels = ax.get_legend_handles_labels()
        handles_new=[Line2D([0],[0]) for h in handles]

        for h,h_new in zip(handles,handles_new):
            h_new.update_from(h)
            h_new.set_marker("")

        if legend_location is None:
            # loc, pos = ['lower center', np.array((0.5, 0))]
            ax.legend(handles_new,labels,fancybox=True, )
        else:
            loc, pos = legend_location
            ax.legend(handles_new,labels,loc=loc, bbox_to_anchor=pos,
                      fancybox=True)
    return f


def get_default(l:List,i:int,default):
    if l is None:
        return default
    else:
        return l[i]


def add_legends(ax, original_labels:List[str], plot_mean:bool, legend_location):
    put_legend=False
    if plot_mean:
        handles, labels = ax.get_legend_handles_labels()
        handles_new =[handles[-1]]
        labels=[labels[-1]]
        put_legend=True
    else:
        if not original_labels is None:
            # Put legend below current axis
            handles, labels = ax.get_legend_handles_labels()
            handles_new = [Line2D([0], [0]) for h in handles]

            for h, h_new in zip(handles, handles_new):
                h_new.update_from(h)
                h_new.set_marker("")
            put_legend = True

    if put_legend:
        if legend_location is None:
            # loc, pos = ['lower center', np.array((0.5, 0))]
            ax.legend(handles_new, labels, fancybox=True)
        else:
            loc, pos = legend_location
            ax.legend(handles_new, labels, loc=loc, bbox_to_anchor=pos,
                      fancybox=True)

def shorten_layer_names(labels:List[str])->List[str]:
    result=[]
    for l in labels:
        i=0
        # get the index of the first letter after the last _
        chars=[str(n) for n in range(9)]+["_"]
        while i<len(l) and l[i] in chars: i+=1
        # remove all chars before the last _
        l=l[i:]

        # if l[1]=="_":
        #     l=l[2:]
        if l.startswith("fc"):
            l="fc"+l[2:]
        if l=="c":
            l="conv"
        if l.endswith("MaxPool2d"):
            l=l[:-9]+"mp"
            result.append(l)
        elif l.endswith("Flatten"):
            l=l[:-7]+"vect"
            result.append(l)
        else:
            result.append(l)
    return result

def get_colors(colors:np.ndarray,n:int)->np.ndarray:
    if colors is None:
        if n==2:
            colors = np.array([[1, 0, 0], [0, 0, 1]])
        else:
            colors = plt.cm.rainbow(np.linspace(0.01, 1, n))
    return colors

def get_dpi(n:int):
    return min(400, max(300, n * 15))

def none_list(v):
    if v == None:
        return None
    else:
        return [v]

def plot_average_activations(result:tm.MeasureResult, label:str=None, title:str="", linestyle:str=None, color:str=None, mark_layers:List[int]=None, ylim=None):
    results = [result]
    colors = none_list(color)
    linestyles = none_list(linestyle)
    labels  = none_list(label)

    return plot_average_activations_same_model(results,title=title,mark_layers=mark_layers,ylim=ylim,labels=labels,linestyles=linestyles,colors=colors)


def scatter_same_model(results:List[tm.MeasureResult], labels:List[str]=None,colors=None,ylim=None):
    n=len(results)
    if n == 0:
        raise ValueError(f"`results` is an empty list.")
    colors=get_colors(colors,n)

    f, ax = plt.subplots(dpi=min(get_dpi(n),200))
    n_layers= len(results[0].layers)
    max_value=0
    for i in range(n_layers):
        y_results = [r.layers[i].flatten().squeeze() for r in results]
        colors_results = [ np.array([c]*r.size)  for r,c in zip(y_results,colors) ]    
        x = np.concatenate([np.zeros_like(r)+i+(j/(n+1))  for j,r in enumerate(y_results) ])
        x += np.random.rand(*x.shape)/(n*3)-0.5
        y = np.concatenate(y_results)
        colors_results = np.concatenate(colors_results)
        
        max_value = max(max_value,y.max())
        # x = np.zeros_like(y)+i+np.random.rand(len(y))*0.9
        ax.scatter(x, y,color=colors_results,marker=",",s=0.5,linewidths=0)

    ax.set_ylabel("Measure")
    ax.set_xlabel("Activation")
    if ylim is None:
        ylim = default_y_lim
    ax.set_ylim(0,max(max_value*1.1,ylim))

    if n_layers < 60:
        tick_labels = results[0].layer_names
        tick_labels = shorten_layer_names(tick_labels)
        #labels = [f"${l}$" for l in labels]
        x = np.arange(n_layers) 
        ax.set_xticks(x)
        ax.set_xticklabels(tick_labels, rotation=90)
        ax.tick_params(axis='both', which='both', length=0)

def plot_average_activations_same_model(results:List[tm.MeasureResult], labels:List[str]=None, title="", linestyles=None, plot_mean=False, colors=None, legend_location=None, mark_layers:List[int]=None, ylim=None,plot_p_values=False):
    if ylim is None:
        ylim = default_y_lim

    n=len(results)
    if n == 0:
        raise ValueError(f"`results` is an empty list.")
    result_layers=np.array([len(r.layer_names) for r in results])
    min_n,max_n = result_layers.min(),result_layers.max()
    assert min_n==max_n,"All results must have the same number of layers."

    colors=get_colors(colors,n)

    f, ax = plt.subplots(dpi=get_dpi(n))
    f.suptitle(title)

    

    if linestyles is None and n <= 4 and plot_mean == False:
        linestyles = ["-", "--", ":", "-."]
    
    mean_and_variance = RunningMeanAndVarianceWelford()
    max_value=0
    for i, result in enumerate(results):
        n_layers= len(result.layers)
        x= np.arange(n_layers)+1
        y= result.per_layer_average()
        max_value = max(max_value,y.max())
        if plot_mean:
            mean_and_variance.update(y)
        label = get_default(labels,i,None)
        linestyle = get_default(linestyles,i,"-")
        color=colors[i, :]
        # if plot_mean:
        #     color*=0.7

        ax.plot(x, y, label=label, linestyle=linestyle, color=color,marker="o",markersize=3)

        if not mark_layers is None:
            x_mark = x[mark_layers]
            y_mark = y[mark_layers]
            ax.plot(x_mark,y_mark,linestyle="",color=color,marker="s")
    if plot_mean:
        x = np.arange(max_n)+1
        y,error=mean_and_variance.mean(),mean_and_variance.std()
        label="μ/σ"
        linestyle="--"
        mean_color = (0,0,0)
        ax.errorbar(x, y,yerr=error, label=label, linewidth=1.5, linestyle=linestyle, color=mean_color)
        if plot_p_values:
            p_values = tukey_same_mean_test(results,alpha=0.001) 
            print("p_values:",p_values)
            for p_value,xi,yi in zip (p_values,x,y):
                p_value_str = f"{p_value*100:.1f}%"
                plt.annotate(p_value_str,(xi,yi), xycoords="data",textcoords="offset points", xytext=(0,15), ha="center",fontsize=5)
    else:
        handles, _ = ax.get_legend_handles_labels()

    ax.set_ylabel("Measure")
    ax.set_xlabel("Activation")
    ax.set_ylim(0,max(max_value*1.1,ylim))

    if max_n < 60:
        tick_labels = results[0].layer_names
        tick_labels = shorten_layer_names(tick_labels)
        #labels = [f"${l}$" for l in labels]
        x = np.arange(max_n) + 1
        ax.set_xticks(x)
        ax.set_xticklabels(tick_labels, rotation=90)
        ax.tick_params(axis='both', which='both', length=0)

    box = ax.get_position()
    ax.set_position([box.x0, box.y0 + box.height * 0.1,
                     box.width, box.height * 0.9])

    add_legends(ax,labels,plot_mean,legend_location)
    return f

